"""Contain input widgets."""

from __future__ import annotations

import asyncio
import logging
from abc import ABCMeta, abstractmethod
from collections import deque
from functools import partial
from math import ceil, floor
from typing import TYPE_CHECKING, cast
from weakref import finalize

from prompt_toolkit.auto_suggest import DynamicAutoSuggest
from prompt_toolkit.buffer import Buffer, ValidationState
from prompt_toolkit.cache import SimpleCache
from prompt_toolkit.completion import Completer, ConditionalCompleter, WordCompleter
from prompt_toolkit.data_structures import Point
from prompt_toolkit.document import Document
from prompt_toolkit.filters import (
    Always,
    Condition,
    Filter,
    FilterOrBool,
    has_focus,
    is_true,
    to_filter,
)
from prompt_toolkit.formatted_text.base import to_formatted_text
from prompt_toolkit.formatted_text.utils import fragment_list_len, fragment_list_width
from prompt_toolkit.key_binding.key_bindings import (
    ConditionalKeyBindings,
    KeyBindings,
    merge_key_bindings,
)
from prompt_toolkit.layout.containers import ConditionalContainer, Float, VSplit, Window
from prompt_toolkit.layout.controls import (
    BufferControl,
    FormattedTextControl,
    UIContent,
    UIControl,
)
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.layout.processors import (
    AfterInput,
    BeforeInput,
    ConditionalProcessor,
    PasswordProcessor,
    Processor,
)
from prompt_toolkit.layout.screen import WritePosition
from prompt_toolkit.layout.utils import explode_text_fragments
from prompt_toolkit.lexers import DynamicLexer, Lexer
from prompt_toolkit.mouse_events import MouseButton, MouseEvent, MouseEventType
from prompt_toolkit.utils import Event
from prompt_toolkit.validation import Validator

from euporie.core.app.current import get_app
from euporie.core.border import InsetGrid
from euporie.core.data_structures import DiBool
from euporie.core.ft.utils import FormattedTextAlign, align
from euporie.core.margins import MarginContainer, ScrollbarMargin
from euporie.core.widgets.decor import Border, Shadow
from euporie.core.widgets.layout import Box, ConditionalSplit

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any, Callable

    from prompt_toolkit.auto_suggest import AutoSuggest
    from prompt_toolkit.buffer import BufferAcceptHandler
    from prompt_toolkit.completion.base import Completer
    from prompt_toolkit.formatted_text.base import (
        AnyFormattedText,
        OneStyleAndTextTuple,
        StyleAndTextTuples,
    )
    from prompt_toolkit.key_binding.key_bindings import (
        KeyBindingsBase,
        NotImplementedOrNone,
    )
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.layout.controls import (
        GetLinePrefixCallable,
        SearchBufferControl,
    )
    from prompt_toolkit.layout.processors import Processor
    from prompt_toolkit.lexers import Lexer

    from euporie.core.border import GridStyle

    OptionalSearchBuffer = (
        SearchBufferControl | Callable[[], SearchBufferControl] | None
    )

log = logging.getLogger(__name__)


class Swatch:
    """An widget which displays a given color."""

    def __init__(
        self,
        color: str | Callable[[], str] = "#FFFFFF",
        width: int = 2,
        height: int = 1,
        style: str = "class:swatch",
        border: GridStyle = InsetGrid,
        show_borders: DiBool | None = None,
    ) -> None:
        """Create a new instance of the color swatch.

        Args:
            color: A function or string which gives color to display
            width: The width of the color swatch (excluding the border)
            height: The height of the color swatch (excluding the border)
            style: Additional style to apply to the color swatch
            border: The grid style to use for the widget's border
            show_borders: Determines which borders should be displayed

        """
        self.color = color
        self.style = style

        self.container = Border(
            Window(char=" ", style=self.get_style, width=width, height=height),
            border=border,
            style=f"{self.style} class:border,inset",
            show_borders=show_borders,
        )

    def get_style(self) -> str:
        """Compute the style for the swatch.."""
        color = self.color() if callable(self.color) else self.color
        return f"{self.style} bg:{color} "

    def __pt_container__(self) -> AnyContainer:
        """Return the swatch container."""
        return self.container


class Button:
    """A clickable button widget."""

    def __init__(
        self,
        text: AnyFormattedText,
        on_click: Callable[[Button], None] | None = None,
        on_mouse_down: Callable[[Button], None] | None = None,
        disabled: FilterOrBool = False,
        width: int | None = None,
        style: str | Callable[[], str] = "class:input",
        border: GridStyle | None = InsetGrid,
        show_borders: DiBool | None = None,
        selected: bool = False,
        key_bindings: KeyBindingsBase | None = None,
        mouse_handler: Callable[[MouseEvent], NotImplementedOrNone] | None = None,
    ) -> None:
        """Create a new button widget instance.

        Args:
            text: The caption for the button.
            on_click: A callback to run when the mouse is released over the button
            on_mouse_down: A callback to run when the mouse is pressed on the button
            disabled: A filter which when evaluated to :py:const:`True` causes the
                widget to be disabled
            width: The width of the button. If :py:const:`None`, the button width is
                determined by the text
            style: A style string or callable style to apply to the button
            border: The grid style to use as the button's border
            show_borders: Determines which borders should be shown
            selected: The selection state of the button
            key_bindings: Additional key_binding to apply to the button
            mouse_handler: A mouse handler for the button. If unset, the default will be
                used, which results in the button behaving like a regular click-button

        """
        self.text = text
        self.on_mouse_down = Event(self, on_mouse_down)
        self.on_click = Event(self, on_click)
        self.disabled = to_filter(disabled)
        self._width = width
        self.style = style
        self.selected = selected
        if key_bindings is not None:
            self.key_bindings: KeyBindingsBase = merge_key_bindings(
                [self.get_key_bindings(), key_bindings]
            )
        else:
            self.key_bindings = self.get_key_bindings()
        self.mouse_handler = mouse_handler or self.default_mouse_handler
        self.window = Window(
            FormattedTextControl(
                self.get_text_fragments,
                key_bindings=self.key_bindings,
                focusable=~self.disabled,
                show_cursor=False,
                style="class:face",
            ),
            style=self.get_style,
            dont_extend_width=True,
            dont_extend_height=True,
        )
        self.container = Box(
            Border(
                self.window,
                border=border,
                show_borders=show_borders,
                style=lambda: f"{self.get_style()} class:border",
            ),
            padding=0,
        )
        self.has_focus = has_focus(self)

    def get_style(self) -> str:
        """Return the style for the button given its current state."""
        style = self.style() if callable(self.style) else self.style
        style = f"{style} class:button"
        if self.selected:
            style = f"{style} class:selection"
        if self.disabled():
            style = f"{style} class:disabled"
        return style

    def get_text_fragments(self) -> StyleAndTextTuples:
        """Return the list of formatted text fragments which define the button."""
        ft = [
            *to_formatted_text(self.text),
        ]
        ft = align(ft, FormattedTextAlign.CENTER, self.width)
        ft = [(style, text, self.mouse_handler) for style, text, *_ in ft]
        return [("[SetMenuPosition]", ""), *ft]

    def get_key_bindings(self) -> KeyBindingsBase:
        """Key bindings for the Button."""
        kb = KeyBindings()

        @kb.add(" ", filter=~self.disabled)
        @kb.add("enter", filter=~self.disabled)
        def _(event: KeyPressEvent) -> None:
            self.on_mouse_down.fire()
            self.on_click.fire()

        return kb

    def default_mouse_handler(self, mouse_event: MouseEvent) -> NotImplementedOrNone:
        """Handle mouse events."""
        app = get_app()
        if self.disabled():
            return NotImplemented
        if mouse_event.button == MouseButton.LEFT:
            if mouse_event.event_type == MouseEventType.MOUSE_DOWN:
                self.selected = True
                # Set global mouse capture
                if (
                    self.window is not None
                    and (render_info := self.window.render_info) is not None
                    and render_info._rowcol_to_yx
                ):
                    y_min, x_min = min(render_info._rowcol_to_yx.values())
                    y_max, x_max = max(render_info._rowcol_to_yx.values())
                    app.mouse_limits = WritePosition(
                        xpos=x_min,
                        ypos=y_min,
                        width=x_max - x_min + 1,
                        height=y_max - y_min + 1,
                    )
                self.on_mouse_down.fire()

                if not self.has_focus():
                    app.layout.focus(self)
                    # Invalidate the app - we don't do it by returning None so the
                    # event can bubble
                    app.invalidate()
                    # We want this event to bubble if unfocused
                    return NotImplemented
                else:
                    return None

            elif mouse_event.event_type == MouseEventType.MOUSE_UP:
                app.mouse_limits = None
                if self.selected:
                    self.selected = False
                    self.on_click.fire()
                return None

            elif mouse_event.event_type == MouseEventType.MOUSE_MOVE:
                # Unselect the button if the mouse is moved outside of the button
                if (info := self.window.render_info) is not None and (
                    info._x_offset + mouse_event.position.x,
                    info._y_offset + mouse_event.position.y,
                ) != app.mouse_position:
                    app.mouse_limits = None
                    self.selected = False
                return None

        app.mouse_limits = None
        self.selected = False
        return NotImplemented

    @property
    def width(self) -> int:
        """The width of the button."""
        if self._width is not None:
            return self._width
        else:
            return fragment_list_width(to_formatted_text(self.text)) + 2

    @width.setter
    def width(self, value: int | None) -> None:
        """Set the width of the button."""
        self._width = value

    def __pt_container__(self) -> AnyContainer:
        """Return the button's container."""
        return self.container


class ToggleableWidget(metaclass=ABCMeta):
    """Base class for toggleable widgets."""

    container: AnyContainer
    on_click: Event
    selected: bool
    disabled: Filter
    key_bindings: KeyBindingsBase | None

    def toggle(self) -> None:
        """Toggle the selected state and trigger the "clicked" callback."""
        self.selected = not self.selected
        self.on_click.fire()

    def mouse_handler(self, mouse_event: MouseEvent) -> NotImplementedOrNone:
        """Focus on mouse down and toggle state on mouse up."""
        if self.disabled():
            return NotImplemented
        elif mouse_event.event_type == MouseEventType.MOUSE_DOWN:
            layout = get_app().layout
            if not layout.has_focus(self):
                app = get_app()
                app.layout.focus(self)
                app.invalidate()
                return NotImplemented
            else:
                return None
        elif mouse_event.event_type == MouseEventType.MOUSE_UP:
            self.toggle()
            return None
        else:
            return NotImplemented

    def _get_key_bindings(self) -> KeyBindingsBase:
        """Key bindings for the Toggler."""
        kb = KeyBindings()

        @kb.add(" ", filter=~self.disabled)
        @kb.add("enter", filter=~self.disabled)
        def _(event: KeyPressEvent) -> None:
            self.toggle()

        if self.key_bindings is None:
            return kb
        else:
            return merge_key_bindings([kb, self.key_bindings])

    def __pt_container__(self) -> AnyContainer:
        """Return the toggler's container."""
        return self.container


class ToggleButton(ToggleableWidget):
    """A toggleable button widget."""

    def __init__(
        self,
        text: AnyFormattedText,
        on_click: Callable[[ToggleButton], None] | None = None,
        width: int | None = None,
        style: str | Callable[[], str] = "class:input",
        border: GridStyle | None = InsetGrid,
        show_borders: DiBool | None = None,
        selected: bool = False,
        disabled: FilterOrBool = False,
        key_bindings: KeyBindingsBase | None = None,
    ) -> None:
        """Create a new toggle-button instance.

        Args:
            text: The text to display on the button (can be a string or callable)
            on_click: A callback to run when the button is clicked
            width: The width of the button
            style: A style for the button
            border: The grid style to use as the button's border
            show_borders: Which borders to display
            selected: The initial selection state of the button
            disabled: A filter which when evaluated to :py:const:`True` causes the
                widget to be disabled
            key_bindings: Additional key_binding to apply to the button
        """
        self.on_click = Event(self, on_click)
        self.style = style
        self.disabled = to_filter(disabled)
        self.key_bindings = key_bindings

        self.button = Button(
            text=text,
            width=width,
            style=style,
            border=border,
            show_borders=show_borders,
            selected=selected,
            key_bindings=self._get_key_bindings(),
            mouse_handler=self.mouse_handler,
            on_click=lambda button: self.on_click.fire(),
            disabled=self.disabled,
        )
        self.container = self.button

    # See mypy issue #4125
    @property  # type: ignore
    def selected(self) -> bool:  # type: ignore
        """Return the selection state of the toggle button."""
        return self.button.selected

    @selected.setter
    def selected(self, value: bool) -> None:
        """Set the selection state of the toggle button."""
        self.button.selected = value


class Checkbox(ToggleableWidget):
    """A toggleable checkbox widget."""

    def _get_text_fragments(self) -> StyleAndTextTuples:
        """Return the text fragments to display."""
        selected_style = "class:selection" if self.selected else ""
        ft = [
            (
                f"[SetCursorPosition] class:checkbox,prefix {selected_style}",
                self.prefix[int(self.selected)],
            ),
            # Add space between the tickbox and the text
            ("", " " if fragment_list_len(self.text) else ""),
            *self.text,
        ]
        return [(style, text, self.mouse_handler) for style, text, *_ in ft]

    def __init__(
        self,
        text: AnyFormattedText = "",
        on_click: Callable[[ToggleableWidget], None] | None = None,
        prefix: tuple[str, str] = ("☐", "☑"),
        style: str = "class:input",
        selected: bool = False,
        disabled: FilterOrBool = False,
        key_bindings: KeyBindingsBase | None = None,
    ) -> None:
        """Create a new checkbox widget instance.

        Args:
            text: The text to display next to the checkbox
            on_click: A callback to run when the checkbox is toggled
            prefix: A tuple of prefix strings to display, representing a checkbox in
                unchecked and checked state respectively
            style: Additional style string to apply to the widget
            selected: The initial selection state of the widget
            disabled: A filter which when evaluated to :py:const:`True` causes the
                widget to be disabled
            key_bindings: Additional key_binding to apply to the button

        """
        self.text = to_formatted_text(text)
        self.on_click = Event(self, on_click)
        self.prefix = prefix
        self.style = style
        self.selected = selected
        self.disabled = to_filter(disabled)
        self.key_bindings = key_bindings
        self.container = Window(
            FormattedTextControl(
                self._get_text_fragments,
                key_bindings=self._get_key_bindings(),
                focusable=~self.disabled,
                show_cursor=False,
            ),
            style=lambda: f"class:checkbox {style}"
            f"{' class:disabled' if self.disabled() else ''}",
            dont_extend_width=True,
            dont_extend_height=True,
        )


class ExpandingBufferControl(BufferControl):
    """A sub-class of :class:`BufferControl` which expands to the available width."""

    def __init__(
        self,
        buffer: Buffer | None = None,
        input_processors: list[Processor] | None = None,
        include_default_input_processors: bool = True,
        lexer: Lexer | None = None,
        preview_search: FilterOrBool = False,
        focusable: FilterOrBool = True,
        search_buffer_control: OptionalSearchBuffer = None,
        menu_position: Callable[[], int | None] | None = None,
        focus_on_click: FilterOrBool = False,
        key_bindings: KeyBindingsBase | None = None,
        expand: FilterOrBool = True,
        on_focus: Callable[[], None] | None = None,
    ) -> None:
        """Add an ``expand`` parameter to the buffer control."""
        super().__init__(
            buffer=buffer,
            input_processors=input_processors,
            include_default_input_processors=include_default_input_processors,
            lexer=lexer,
            preview_search=preview_search,
            focusable=focusable,
            search_buffer_control=search_buffer_control,
            menu_position=menu_position,
            focus_on_click=focus_on_click,
            key_bindings=key_bindings,
        )
        self.expand = to_filter(expand)
        self.on_focus = on_focus

    def preferred_width(self, max_available_width: int) -> int | None:
        """Enure text box expands to available width.

        Args:
            max_available_width: The maximum available width

        Returns:
            The desired width, which is the maximum available
        """
        if self.expand():
            return max_available_width
        else:
            return None

    def mouse_handler(self, mouse_event: MouseEvent) -> NotImplementedOrNone:
        """Optionally call focus handler when focused."""
        layout = get_app().layout
        was_focused = layout.current_control == self
        result = super().mouse_handler(mouse_event)
        is_focused = layout.current_control == self
        if not was_focused and is_focused and callable(self.on_focus):
            self.on_focus()
        return result


class Text:
    """A text input widget."""

    def __init__(
        self,
        text: str = "",
        style: str = "class:input",
        height: int = 1,
        min_height: int = 1,
        multiline: FilterOrBool = False,
        expand: FilterOrBool = True,
        width: int | None = None,
        completer: Completer | None = None,
        options: list[str] | Callable[[], list[str]] | None = None,
        show_borders: DiBool | None = None,
        on_text_changed: Callable[[Buffer], None] | None = None,
        validation: Callable[[str], bool] | None = None,
        accept_handler: BufferAcceptHandler | None = None,
        placeholder: str | None = None,
        lexer: Lexer | None = None,
        input_processors: Sequence[Processor] | None = None,
        disabled: FilterOrBool = False,
        password: FilterOrBool = False,
        wrap_lines: FilterOrBool = False,
        prompt: AnyFormattedText | None = None,
        on_focus: Callable[[], None] | None = None,
        complete_while_typing: FilterOrBool = True,
        auto_suggest: AutoSuggest | None = None,
    ) -> None:
        """Create a new text widget instance.

        Args:
            text: The default value of the text widget
            style: Additional style string to apply to the widget
            height: The height of the widget, excluding borders
            min_height: The minimum height of the widget, excluding borders
            multiline: Whether the text box accepts multiple lines of text
            expand: Determines if the text input should expand to the available width
            width: The width of the text input. If :py:const:`None`, the widget will
                expand to the available width
            completer: A completer to use, an alternative to ``options``
            options: A list of permitted values
            show_borders: Which borders to display. By default, all borders are shown
            on_text_changed: A callback to run when the text is changed
            validation: A callable used to validate the input
            accept_handler: A callable which run when the input is accepted (when the
                :kbd:`Enter` key is pressed on a non-multiline input)
            placeholder: Text to display when nothing has been entered
            lexer: :class:`~prompt_toolkit.lexers.Lexer` instance for syntax
                highlighting
            input_processors: Additional input processors to apply to the text-area
            disabled: A filter which when evaluated to :py:const:`True` causes the
                widget to be disabled
            password: A filter to determine if the text input is a password field
            wrap_lines: Whether to wrap lines wider than the text area
            prompt: Text to display before the input
            on_focus: Function to call when the buffer gains focus
            complete_while_typing: Whether to show completions while typing
            auto_suggest: Auto-suggestion behavior for the text input
        """
        self.style = style
        self.options = options or []
        self.disabled = to_filter(disabled)
        self.placeholder = placeholder
        self.wrap_lines = wrap_lines
        self.complete_while_typing = complete_while_typing
        self.auto_suggest = auto_suggest
        self.lexer = lexer

        self.buffer = Buffer(
            document=Document(str(text), 0),
            multiline=multiline,
            read_only=self.disabled,
            completer=completer
            or ConditionalCompleter(
                WordCompleter(self.options),
                filter=Condition(lambda: bool(self.options)),
            ),
            complete_while_typing=Condition(
                lambda: is_true(self.complete_while_typing)
            ),
            validator=Validator.from_callable(validation) if validation else None,
            validate_while_typing=validation is not None,
            auto_suggest=DynamicAutoSuggest(lambda: self.auto_suggest),
            accept_handler=accept_handler,
        )

        self.control = ExpandingBufferControl(
            buffer=self.buffer,
            lexer=DynamicLexer(lambda: self.lexer),
            input_processors=[
                BeforeInput(prompt, style="class:text-area.prompt"),
                ConditionalProcessor(
                    AfterInput(
                        lambda: self.placeholder, style="class:text,placeholder"
                    ),
                    filter=Condition(
                        lambda: self.placeholder is not None and self.buffer.text == ""
                    ),
                ),
                ConditionalProcessor(
                    processor=PasswordProcessor(), filter=to_filter(password)
                ),
                *(input_processors or []),
            ],
            focusable=~self.disabled,
            focus_on_click=~self.disabled,
            expand=expand,
            on_focus=on_focus,
        )

        self.window = Window(
            height=Dimension(min=min_height, max=height, preferred=height),
            width=width,
            content=self.control,
            wrap_lines=Condition(lambda: is_true(self.wrap_lines)),
            style=f"{style} class:text,text-area",
        )

        if on_text_changed:
            self.buffer.on_text_changed += on_text_changed
        if validation:
            self.buffer.validate_while_typing = Always()
        self.container = Border(
            VSplit(
                [
                    self.window,
                    ConditionalContainer(
                        MarginContainer(ScrollbarMargin(), target=self.window),
                        filter=to_filter(multiline),
                    ),
                ]
            ),
            border=InsetGrid,
            style=self.border_style,
            show_borders=show_borders,
        )

    def border_style(self) -> str:
        """Calculate the style to apply to the widget's border."""
        if self.buffer.validation_state == ValidationState.INVALID:
            return f"{self.style} class:text,border,invalid"
        else:
            return f"{self.style} class:text,border"

    @property
    def text(self) -> str:
        """Return the input's text value."""
        return self.buffer.text

    @text.setter
    def text(self, value: str) -> None:
        """Set the input's text value."""
        self.buffer.text = value

    def __pt_container__(self) -> AnyContainer:
        """Return the widget's container."""
        return self.container


class Label:
    """A label widget which displays rich text."""

    def __init__(
        self,
        value: AnyFormattedText,
        style: str | Callable[[], str] = "class:input",
        html: FilterOrBool = False,
    ) -> None:
        """Create a new label widget instance.

        Args:
            value: The value to display
            style: Additional style to apply to the widget
            html: Whether to render the label as HTML
        """
        self.value = value
        self.control = FormattedTextControl(self.get_value, focusable=False)
        self.html = to_filter(html)
        self.container = Window(
            self.control,
            style=style,
            dont_extend_width=True,
            dont_extend_height=True,
        )

    def get_value(self) -> AnyFormattedText:
        """Return the current value of the label, converting to formatted text."""
        value = self.value
        data = value() if callable(value) else value
        if self.html():
            from prompt_toolkit.formatted_text.utils import to_plain_text

            from euporie.core.ft.html import HTML

            return HTML(to_plain_text(data), collapse_root_margin=True, fill=False)
        return data

    def __pt_container__(self) -> AnyContainer:
        """Return the widget's container."""
        return self.container


class LabelledWidget:
    """A widget which applies a label to another widget."""

    def __init__(
        self,
        body: AnyContainer,
        label: AnyFormattedText,
        style: str = "class:input",
        vertical: FilterOrBool = False,
        html: FilterOrBool = False,
    ) -> None:
        """Create a new labelled widget instance.

        Args:
            body: The widget to label
            label: The label text to apply
            style: Additional style string to apply to the label
            vertical: Determines if the labelled widget should be oriented vertically
            html: Whether to render the label as HTML
        """
        self.body = body
        self.vertical = to_filter(vertical)
        self._html = to_filter(html)
        self.label = Label(label, style=style, html=self.html)
        padding_left = padding_right = lambda: None if self.vertical() else 0
        padding_top = padding_bottom = lambda: 0 if self.vertical() else None
        self.container = ConditionalSplit(
            self.vertical,
            [
                ConditionalContainer(
                    Box(
                        self.label,
                        padding_top=padding_top,
                        padding_right=padding_right,
                        padding_bottom=padding_bottom,
                        padding_left=padding_left,
                    ),
                    filter=Condition(
                        lambda: bool(fragment_list_len(to_formatted_text(label)))
                    ),
                ),
                self.body,
            ],
            padding=1,
            style=style,
        )

    @property
    def html(self) -> Filter:
        """Get the HTML filter value."""
        return self._html

    @html.setter
    def html(self, value: FilterOrBool) -> None:
        """Set the HTML filter value."""
        self._html = to_filter(value)
        self.label.html = self._html

    def __pt_container__(self) -> AnyContainer:
        """Return the labelled widget container."""
        return self.container


class ProgressControl(UIControl):
    """A control which draws a progress-bar."""

    hchars = ("", "▏", "▎", "▍", "▌", "▋", "▊", "▉", "█")
    vchars = ("", "▁", "▂", "▃", "▄", "▅", "▆", "▇", "█")

    def __init__(
        self,
        start: float | int = 0,
        stop: float | int = 100,
        step: float | int = 1,
        value: float | int = 0,
        vertical: FilterOrBool = False,
    ) -> None:
        """Create a new progress-bar control instance.

        Args:
            start: The lowest value of the progress-bar
            stop: The highest value of the progress-bar
            step: The size of each progress interval
            value: The initial value to display
            vertical: Determines if the progress-bar should be drawn in a vertical
                orientation
        """
        self.start = start
        self.stop = stop
        self.step = step
        self.value = value
        self.vertical = to_filter(vertical)

        self._content_cache: SimpleCache = SimpleCache(maxsize=50)

    def preferred_width(self, max_available_width: int) -> int | None:
        """Determine the width of the progress-bar depending on its orientation."""
        return 1 if self.vertical() else max_available_width

    def preferred_height(
        self,
        width: int,
        max_available_height: int,
        wrap_lines: bool,
        get_line_prefix: GetLinePrefixCallable | None,
    ) -> int | None:
        """Determine the height of the progress-bar depending on its orientation."""
        return min(max_available_height, 10) if self.vertical() else 1

    def render(self, width: int, height: int) -> list[StyleAndTextTuples]:
        """Render the progressbar at a given size as lines of formatted text."""
        vertical = self.vertical()
        length = height if vertical else width
        size = int(self.value / (self.stop - self.start) * length * 8) / 8
        remainder = int(size % 1 * 8)
        chars = self.vchars if vertical else self.hchars

        ft: StyleAndTextTuples = [
            ("class:progress", chars[8] * int(size)),
            ("class:progress", chars[remainder]),
            ("class:progress", " " * int(length - size)),
        ]

        if vertical:
            return [[x] for x in explode_text_fragments(ft)][::-1]
        else:
            return [ft]

    def create_content(self, width: int, height: int) -> UIContent:
        """Get or render content for a given output size."""

        def get_content() -> UIContent:
            fragment_lines = self.render(width, height)

            return UIContent(
                get_line=lambda i: fragment_lines[i],
                line_count=len(fragment_lines),
                show_cursor=False,
            )

        key = (width, self.start, self.stop, self.step, self.value)
        return self._content_cache.get(key, get_content)


class Progress:
    """A progress-bar widget."""

    def __init__(
        self,
        start: float | int = 0,
        stop: float | int = 100,
        step: float | int = 1,
        value: float | int = 0,
        vertical: FilterOrBool = False,
        style: str | Callable[[], str] = "class:input",
    ) -> None:
        """Create a new progress-bar widget instance.

        Args:
            start: The lowest permitted value
            stop: The highest permitted value
            step: The interval between permitted values
            value: The initial value
            vertical: Determines if the progress-bar should be drawn in a vertical
                orientation
            style: Additional style string or callable to apply to the widget
        """
        self.style = style
        self.vertical = to_filter(vertical)
        self.control = ProgressControl(
            start=start,
            stop=stop,
            step=step,
            value=value,
            vertical=self.vertical,
        )
        self.container = Box(
            Border(
                Window(
                    self.control,
                    style=self.add_style("class:progress"),
                    dont_extend_width=self.vertical,
                    # dont_extend_height=~self.vertical,
                    height=lambda: None if self.vertical() else 1,
                    width=lambda: 1 if self.vertical() else None,
                ),
                border=InsetGrid,
                style=self.add_style("class:progress,border"),
            ),
            # width=lambda: 1 if self.vertical() else None,
        )

    @property
    def value(self) -> float | int:
        """Return the current value of the proegress-bar."""
        return self.control.value

    @value.setter
    def value(self, value: float | int) -> None:
        """Set the current value of the proegress-bar."""
        self.control.value = value

    def add_style(self, extra: str) -> Callable[[], str]:
        """Add an additional style to the widget's base style."""

        def _style() -> str:
            if callable(self.style):
                return f"{self.style()} {extra}"
            else:
                return f"{self.style} {extra}"

        return _style

    def __pt_container__(self) -> AnyContainer:
        """Return the progress-bar's container."""
        return self.container


class SizedMask(dict[int, bool]):
    """Mask with restricted number of True items."""

    def __init__(self, size: int | None = None) -> None:
        """Initialize a new sized default dict."""
        self.size = size
        self._keys: deque[int] = deque()

    def clear(self) -> None:
        """Clear the dict's items."""
        self._keys.clear()
        super().clear()

    def __setitem__(self, key: int, value: bool) -> None:
        """Remove the oldest key when the size is exceeded."""
        if value:
            super().__setitem__(key, value)
        else:
            if key in self:
                del self[key]
            if key in self._keys:
                self._keys.remove(key)

        if self.size:
            if len(self._keys) >= self.size:
                key_to_remove = self._keys.popleft()
                if key_to_remove in self:
                    del self[key_to_remove]
            if value:
                self._keys.append(key)

    def __missing__(self, key: int) -> bool:
        """Return ``False`` for unknown items."""
        return False


class SelectableWidget(metaclass=ABCMeta):
    """Base class for widgets where one or more items can be selected."""

    def __init__(
        self,
        options: list[Any],
        labels: Sequence[AnyFormattedText] | None = None,
        index: int | None = None,
        indices: list[int] | None = None,
        n_values: int | None = None,
        multiple: FilterOrBool = False,
        max_count: int | None = None,
        on_change: Callable[[SelectableWidget], None] | None = None,
        style: str | Callable[[], str] = "class:input",
        disabled: FilterOrBool = False,
    ) -> None:
        """Create a new selectable widget instance.

        Args:
            options: List of permitted values
            labels: Optional list of labels for each permitted value
            index: The index of the initially selected single value
            indices: List of indices of the initially selected values
            n_values: The number of values which are selectable
            multiple: Determines whether multiple values can be selected
            max_count: The maximum number of selectable items
            on_change: Callback which is run when the selection changes
            style: Additional style to apply to the widget
            disabled: A filter which when evaluated to :py:const:`True` causes the
                widget to be disabled
        """
        self.options = options
        self.labels: list[AnyFormattedText] = list(
            labels or (str(option) for option in options)
        )
        self._selected = SizedMask(size=max_count)
        self.multiple = to_filter(multiple)

        if index is None and indices is None:
            indices = [] if self.multiple() else [0]
        if indices is None:
            indices = [index or 0]
        self.indices = indices
        self.n_values = n_values
        self.on_change = Event(self, on_change)
        self._style = style
        self.disabled = to_filter(disabled)

        self.hovered: int | None = index
        self.container = self.load_container()
        self.has_focus = has_focus(self)

    @abstractmethod
    def load_container(self) -> AnyContainer:
        """Abstract method for loading the widget's container."""
        ...

    @property
    def style(self) -> str:
        """Return the widget's style."""
        style = self._style() if callable(self._style) else self._style
        if self.disabled():
            style = f"{style} class:disabled"
        return style

    @style.setter
    def style(self, value: str | Callable[[], str]) -> None:
        """Set the widget's style."""
        self._style = value

    def hover_rel(self, rel: int) -> None:
        """Hover an index relative to the current hovered index."""
        self.hovered = max(0, min((self.hovered or 0) + rel, len(self.options) - 1))

    def select_rel(self, rel: int) -> None:
        """Select an index relative to the current index."""
        self.toggle_item(max(0, min((self.index or 0) + rel, len(self.options) - 1)))

    def key_bindings(self) -> KeyBindingsBase:
        """Key bindings for the selectable widget."""
        kb = KeyBindings()

        @kb.add("home", filter=~self.multiple)
        def _(event: KeyPressEvent) -> None:
            self.toggle_item(0)

        @kb.add("up", filter=~self.multiple)
        def _(event: KeyPressEvent) -> None:
            self.select_rel(-1)

        @kb.add("down", filter=~self.multiple)
        def _(event: KeyPressEvent) -> None:
            self.select_rel(+1)

        @kb.add("end", filter=~self.multiple)
        def _(event: KeyPressEvent) -> None:
            self.toggle_item(len(self.options) - 1)

        @kb.add("enter", filter=self.multiple)
        @kb.add(" ", filter=self.multiple)
        def _(event: KeyPressEvent) -> None:
            if self.hovered is not None:
                self.toggle_item(self.hovered)

        @kb.add("home", filter=self.multiple)
        def _(event: KeyPressEvent) -> None:
            self.hovered = 0

        @kb.add("up", filter=self.multiple)
        def _(event: KeyPressEvent) -> None:
            self.hover_rel(-1)

        @kb.add("down", filter=self.multiple)
        def _(event: KeyPressEvent) -> None:
            self.hover_rel(+1)

        @kb.add("end", filter=self.multiple)
        def _(event: KeyPressEvent) -> None:
            self.hovered = len(self.options) - 1

        return ConditionalKeyBindings(kb, filter=~self.disabled)

    def toggle_item(self, index: int) -> None:
        """Toggle the selection status of the option at a given index."""
        if not self.multiple() and any(self._selected.values()):
            self._selected.clear()
        self._selected[index] = not self._selected[index]
        self.hovered = index
        self.on_change.fire()

    @property
    def mask(self) -> list[bool]:
        """Get mask of selected options."""
        return [self._selected[i] for i in range(len(self.options))]

    @property
    def index(self) -> int | None:
        """Return the first selected index."""
        return next((x for x in self.indices), None)

    @index.setter
    def index(self, value: int) -> None:
        """Set the selected indices to a single value."""
        self.indices = [value]

    @property
    def indices(self) -> list[int]:
        """Return a list of the selected indices."""
        output = [i for i, m in enumerate(self.mask) if m]
        if self.n_values is not None and len(output) < self.n_values:
            output = (output * self.n_values)[: self.n_values]
        return output

    @indices.setter
    def indices(self, values: tuple[int]) -> None:
        """Set the selected indices."""
        self._selected.clear()
        for i in range(len(self.options)):
            self._selected[i] = i in values

    @property
    def value(self) -> Any:
        """Return the selected value."""
        if self.options:
            return self.options[self.index or 0]
        else:
            return None

    @property
    def values(self) -> list[Any]:
        """Return a list of the selected values."""
        if self.options:
            return [self.options[i] for i in self.indices]
        else:
            return [None for i in self.indices]

    def mouse_handler(self, i: int, mouse_event: MouseEvent) -> NotImplementedOrNone:
        """Handle mouse events."""
        if self.disabled():
            return NotImplemented
        if mouse_event.event_type == MouseEventType.MOUSE_MOVE:
            self.hovered = i
            return None
        elif mouse_event.event_type == MouseEventType.MOUSE_DOWN:
            if not self.has_focus():
                app = get_app()
                get_app().layout.focus(self)
                app.invalidate()
                # We want this event to bubble if unfocused
                return NotImplemented
            else:
                return None
        elif mouse_event.event_type == MouseEventType.MOUSE_UP:
            self.toggle_item(i)
            return None
        elif mouse_event.event_type == MouseEventType.SCROLL_UP:
            if self.multiple():
                hovered = max(0, min((self.hovered or 0) - 1, len(self.options) - 1))
                if self.hovered == hovered:
                    return NotImplemented
                else:
                    self.hovered = hovered
                    return None
            else:
                self.toggle_item(
                    item := max(0, min((self.index or 0) - 1, len(self.options) - 1))
                )
                return NotImplemented if item == 0 else None
        elif mouse_event.event_type == MouseEventType.SCROLL_DOWN:
            if self.multiple():
                hovered = max(0, min((self.hovered or 0) + 1, len(self.options) - 1))
                if self.hovered == hovered:
                    return NotImplemented
                else:
                    self.hovered = hovered
                    return None
            else:
                self.toggle_item(
                    item := max(0, min((self.index or 0) + 1, len(self.options) - 1))
                )
                return NotImplemented if item == len(self.options) - 1 else None
            return NotImplemented
        else:
            return NotImplemented

    def __pt_container__(self) -> AnyContainer:
        """Return the widget's container."""
        return self.container


class NavigableFormattedTextControl(FormattedTextControl):
    """Formatted text control where the cursor can be moved by scrolling."""

    def __init__(
        self,
        text: AnyFormattedText = "",
        style: str = "",
        focusable: FilterOrBool = False,
        key_bindings: KeyBindingsBase | None = None,
        show_cursor: bool = True,
        modal: bool = False,
        get_cursor_position: Callable[[], Point | None] | None = None,
        move_cursor_up: Callable[[], None] | None = None,
        move_cursor_down: Callable[[], None] | None = None,
    ) -> None:
        """Initialize control with addition of cursor movement functions."""
        super().__init__(
            text,
            style,
            focusable,
            key_bindings,
            show_cursor,
            modal,
            get_cursor_position,
        )
        self._move_cursor_up = move_cursor_up
        self._move_cursor_down = move_cursor_down

    def move_cursor_down(self) -> None:
        """Request to move the cursor down.

        This happens when scrolling down and the cursor is completely at the
        top.
        """
        if callable(self._move_cursor_down):
            self._move_cursor_down()

    def move_cursor_up(self) -> None:
        """Request to move the cursor up."""
        if callable(self._move_cursor_up):
            self._move_cursor_up()


class Select(SelectableWidget):
    """A select widget, which allows one or more items to be selected from a list."""

    def __init__(
        self,
        options: list[Any],
        labels: Sequence[AnyFormattedText] | None = None,
        index: int | None = None,
        indices: list[int] | None = None,
        n_values: int | None = None,
        multiple: FilterOrBool = False,
        max_count: int | None = None,
        on_change: Callable[[SelectableWidget], None] | None = None,
        style: str | Callable[[], str] = "class:input,select",
        rows: int | None = 3,
        prefix: tuple[str, str] = ("", ""),
        border: GridStyle | None = InsetGrid,
        show_borders: DiBool | None = None,
        disabled: FilterOrBool = False,
        dont_extend_width: FilterOrBool = True,
        dont_extend_height: FilterOrBool = True,
    ) -> None:
        """Create a new select widget instance.

        Args:
            options: List of permitted values
            labels: Optional list of labels for each permitted value
            index: The index of the initially selected single value
            indices: List of indices of the initially selected values
            n_values: The number of values which are selectable
            multiple: Determines whether multiple values can be selected
            max_count: The maximum number of selectable items
            on_change: Callback which is run when the selection changes
            style: Additional style to apply to the widget
            rows: The number of rows of options to display
            prefix: A prefix to add to each row
            border: The grid style to use for the widget's border
            show_borders: Which borders to display
            disabled: A filter which when evaluated to :py:const:`True` causes the
                widget to be disabled
            dont_extend_width: When ``True``, don't take up more width then the
                preferred width reported by the control.
            dont_extend_height: When ``True``, don't take up more width then the
                preferred height reported by the control.

        """
        self.rows = rows
        self.prefix = prefix
        self.border = border
        self.show_borders = show_borders
        self.dont_extend_width = dont_extend_width
        self.dont_extend_height = dont_extend_height
        super().__init__(
            options=options,
            labels=labels,
            index=index,
            indices=indices,
            multiple=multiple,
            max_count=max_count,
            on_change=on_change,
            style=style,
            disabled=disabled,
        )

    def text_fragments(self) -> StyleAndTextTuples:
        """Create a list of formatted text fragments to display."""
        ft: StyleAndTextTuples = []
        labels = self.labels
        if labels:
            max_width = max(fragment_list_width(to_formatted_text(x)) for x in labels)
        else:
            max_width = 1
        for i, label in enumerate(labels):
            label = to_formatted_text(label)
            label = align(label, FormattedTextAlign.LEFT, width=max_width)
            style = "class:selection" if self.mask[i] else ""
            if self.hovered == i and self.multiple() and self.has_focus():
                style += " class:hovered"
            ft_option = [
                ("class:prefix", self.prefix[self.mask[i]]),
                ("", " "),
                *label,
                ("", " "),
                ("", "\n"),
            ]
            handler = cast(
                "Callable[[MouseEvent], None]", partial(self.mouse_handler, i)
            )
            ft += [
                (f"{fragment_style} {style}", text, handler)
                for fragment_style, text, *_ in ft_option
            ]
        if ft:
            ft.pop()
        return ft

    def load_container(self) -> AnyContainer:
        """Load the widget's container."""
        return Box(
            Border(
                VSplit(
                    [
                        window := Window(
                            NavigableFormattedTextControl(
                                self.text_fragments,
                                focusable=~self.disabled,
                                show_cursor=False,
                                key_bindings=self.key_bindings(),
                                get_cursor_position=lambda: Point(
                                    x=0, y=self.hovered or 0
                                )
                                if self.multiple()
                                else Point(x=0, y=self.index or 0),
                                move_cursor_up=lambda: (
                                    self.hover_rel
                                    if self.multiple()
                                    else self.select_rel
                                )(-1),
                                move_cursor_down=lambda: (
                                    self.hover_rel
                                    if self.multiple()
                                    else self.select_rel
                                )(+1),
                            ),
                            height=lambda: self.rows,
                            dont_extend_width=self.dont_extend_width,
                            dont_extend_height=self.dont_extend_height,
                            style=f"class:face"
                            f"{' class:disabled' if self.disabled() else ''}",
                        ),
                        MarginContainer(ScrollbarMargin(), target=window),
                    ]
                ),
                border=self.border,
                show_borders=self.show_borders,
                style="class:inset,border",
            ),
            padding=0,
            style=self.style,
        )


class Dropdown(SelectableWidget):
    """A dropdown widget, allowing selection of an item from a menu of options."""

    def __init__(
        self,
        options: list[Any],
        labels: Sequence[AnyFormattedText] | None = None,
        index: int | None = None,
        indices: list[int] | None = None,
        n_values: int | None = None,
        multiple: FilterOrBool = False,
        max_count: int | None = None,
        on_change: Callable[[SelectableWidget], None] | None = None,
        style: str | Callable[[], str] = "class:input",
        arrow: str = "⯆",
        disabled: FilterOrBool = False,
    ) -> None:
        """Create a new drop-down widget instance.

        Args:
            options: List of permitted values
            labels: Optional list of labels for each permitted value
            index: The index of the initially selected single value
            indices: List of indices of the initially selected values
            n_values: The number of values which are selectable
            multiple: Determines whether multiple values can be selected
            max_count: The maximum number of selectable items
            on_change: Callback which is run when the selection changes
            style: Additional style to apply to the widget
            arrow: The character to use for the dropdown arrow
            disabled: A filter which when evaluated to :py:const:`True` causes the
                widget to be disabled
        """
        self.menu_visible: bool = False
        self.arrow: str = arrow
        super().__init__(
            options=options,
            labels=labels,
            index=index,
            indices=indices,
            multiple=multiple,
            max_count=max_count,
            on_change=on_change,
            style=style,
            disabled=disabled,
        )

        self.menu = Float(
            ConditionalContainer(
                Shadow(
                    Window(
                        FormattedTextControl(
                            self.menu_fragments, focusable=self.disabled
                        ),
                        style=f"class:dropdown,dropdown.menu {self.style}",
                    )
                ),
                filter=Condition(lambda: self.menu_visible) & self.has_focus,
            ),
            xcursor=True,
            ycursor=True,
        )
        menu_name = f"dropdown-menu-{hash(self)}"
        app = get_app()
        app.menus[menu_name] = self.menu

        def _cleanup_menu() -> None:
            del app.menus[menu_name]

        finalize(self, _cleanup_menu)

    def load_container(self) -> AnyContainer:
        """Load the widget's container."""
        self.button = ToggleButton(
            self.button_text,
            on_click=self.toggle_menu,
            style=f"class:dropdown {self.style}",
            key_bindings=self.key_bindings(),
            disabled=self.disabled,
        )
        return self.button

    def button_text(self) -> StyleAndTextTuples:
        """Return the text to display on the button."""
        labels = [to_formatted_text(x) for x in self.labels] or [
            to_formatted_text([("", "")])
        ]
        max_width = max(fragment_list_width(x) for x in labels)
        ft = align(
            to_formatted_text(labels[self.index or 0]),
            FormattedTextAlign.LEFT,
            max_width,
        )
        ft += [("", " " if self.arrow else ""), ("class:arrow", self.arrow)]
        return ft

    def toggle_menu(self, button: ToggleButton) -> None:
        """Show or hide the menu."""
        self.menu_visible = not self.menu_visible
        self.hovered = self.index
        get_app().invalidate()

    def menu_fragments(self) -> StyleAndTextTuples:
        """Return formatted text fragment to display in the menu."""
        ft: StyleAndTextTuples = []
        labels = [to_formatted_text(x) for x in self.labels]
        max_width = max(fragment_list_width(x) for x in labels)
        for i, label in enumerate(labels):
            # Pad each option
            formatted_label = align(label, FormattedTextAlign.LEFT, width=max_width + 2)
            item_style = "class:hovered" if i == self.hovered else ""
            handler = partial(self.mouse_handler, i)
            ft.extend(
                [
                    (item_style, " ", handler),
                    *[
                        (f"{item_style} {style}", text, handler)
                        for style, text, *_ in formatted_label
                    ],
                    (item_style, " ", handler),
                    ("", "\n"),
                ]
            )
        # Remove the last newline
        ft.pop()
        return ft

    def mouse_handler(self, i: int, mouse_event: MouseEvent) -> None:
        """Handle mouse events."""
        if self.disabled():
            return
        super().mouse_handler(i, mouse_event)
        if mouse_event.event_type == MouseEventType.MOUSE_UP:
            self.menu_visible = False
            self.button.selected = False

    def key_bindings(self) -> KeyBindingsBase:
        """Return key-bindings for the drop-down widget."""
        menu_visible = Condition(lambda: self.menu_visible)
        kb = KeyBindings()

        @kb.add("enter", filter=menu_visible)
        @kb.add(" ", filter=menu_visible)
        def _(event: KeyPressEvent) -> None:
            self.index = self.hovered
            self.menu_visible = False
            self.button.selected = False
            self.on_change.fire()

        @kb.add("escape", filter=menu_visible)
        def _(event: KeyPressEvent) -> None:
            self.menu_visible = False
            self.button.selected = False

        @kb.add("home", filter=menu_visible)
        def _(event: KeyPressEvent) -> None:
            self.hovered = 0

        @kb.add("up", filter=menu_visible)
        def _(event: KeyPressEvent) -> None:
            self.hovered = max(0, min((self.hovered or 0) - 1, len(self.options) - 1))

        @kb.add("down", filter=menu_visible)
        def _(event: KeyPressEvent) -> None:
            self.hovered = max(0, min((self.hovered or 0) + 1, len(self.options) - 1))

        @kb.add("end", filter=menu_visible)
        def _(event: KeyPressEvent) -> None:
            self.hovered = len(self.options) - 1

        @kb.add("home", filter=~menu_visible)
        def _(event: KeyPressEvent) -> None:
            self.index = 0

        @kb.add("up", filter=~menu_visible)
        def _(event: KeyPressEvent) -> None:
            self.index = max(0, min((self.index or 0) - 1, len(self.options) - 1))

        @kb.add("down", filter=~menu_visible)
        def _(event: KeyPressEvent) -> None:
            self.index = max(0, min((self.index or 0) + 1, len(self.options) - 1))

        @kb.add("end", filter=~menu_visible)
        def _(event: KeyPressEvent) -> None:
            self.index = len(self.options) - 1

        return ConditionalKeyBindings(kb, filter=~self.disabled)


class ToggleButtons(SelectableWidget):
    """A widget where an option is selected using mutually exclusive toggle-buttons."""

    def __init__(
        self,
        options: list[Any],
        labels: Sequence[AnyFormattedText] | None = None,
        index: int | None = None,
        indices: list[int] | None = None,
        n_values: int | None = None,
        multiple: FilterOrBool = False,
        max_count: int | None = None,
        on_change: Callable[[SelectableWidget], None] | None = None,
        style: str | Callable[[], str] = "class:input",
        border: GridStyle | None = InsetGrid,
        disabled: FilterOrBool = False,
        vertical: FilterOrBool = False,
    ) -> None:
        """Create a new select widget instance.

        Args:
            options: List of permitted values
            labels: Optional list of labels for each permitted value
            index: The index of the initially selected single value
            indices: List of indices of the initially selected values
            n_values: The number of values which are selectable
            multiple: Determines whether multiple values can be selected
            max_count: The maximum number of selectable items
            on_change: Callback which is run when the selection changes
            style: Additional style to apply to the widget
            border: The grid style to use for the widget's border
            disabled: A filter which when evaluated to :py:const:`True` causes the
                widget to be disabled
            vertical: Determines if the toggle buttons should be arranged vertically
        """
        self.border = border
        self.disabled = to_filter(disabled)
        self.vertical = to_filter(vertical)
        super().__init__(
            options=options,
            labels=labels,
            index=index,
            indices=indices,
            multiple=multiple,
            max_count=max_count,
            on_change=on_change,
            style=style,
            disabled=self.disabled,
        )

    def load_container(self) -> AnyContainer:
        """Load the widget's container."""
        if self.options:
            if self.vertical():
                show_borders_values = [
                    DiBool(False, True, False, True) for _ in self.options
                ]
                show_borders_values[0] = show_borders_values[0]._replace(top=True)
                show_borders_values[-1] = show_borders_values[-1]._replace(bottom=True)
            else:
                show_borders_values = [
                    DiBool(True, False, True, False) for _ in self.options
                ]
                show_borders_values[0] = show_borders_values[0]._replace(left=True)
                show_borders_values[-1] = show_borders_values[-1]._replace(right=True)
        else:
            show_borders_values = []
        self.buttons = [
            ToggleButton(
                text=label,
                selected=selected,
                on_click=partial(lambda index, button: self.toggle_item(index), i),
                border=self.border,
                show_borders=show_borders,
                style=self.get_button_style(i),
                disabled=self.disabled,
            )
            for i, (label, selected, show_borders) in enumerate(
                zip(
                    self.labels,
                    self.mask,
                    show_borders_values,
                )
            )
        ]
        self.on_change += self.update_buttons
        return ConditionalSplit(
            self.vertical,
            self.buttons,
            style="class:toggle-buttons",
            key_bindings=self.key_bindings(),
        )

    def get_button_style(self, index: int) -> Callable[[], str]:
        """Return the current button style."""

        def _get_style() -> str:
            style = self.style() if callable(self.style) else self.style
            if self.hovered == index:
                return f"{style} class:hovered"
            else:
                return style

        return _get_style

    @property
    def index(self) -> int | None:
        """Return the first selected index."""
        return next((x for x in self.indices), None)

    @index.setter
    def index(self, value: int) -> None:
        """Set the selected indices to a single value."""
        self.indices = [value]
        self.update_buttons(self)

    def update_buttons(self, widget: SelectableWidget | None = None) -> None:
        """Set the toggle buttons' selection state when the selected index changes."""
        for i, selected in enumerate(self.mask):
            self.buttons[i].selected = selected

    def key_bindings(self) -> KeyBindingsBase:
        """Return key-bindings for the drop-down widget."""
        multiple = self.multiple
        kb = KeyBindings()

        @kb.add("left", filter=~multiple)
        def _(event: KeyPressEvent) -> None:
            self.index = max(0, min((self.index or 0) - 1, len(self.options) - 1))

        @kb.add("right", filter=~multiple)
        def _(event: KeyPressEvent) -> None:
            self.index = max(0, min((self.index or 0) + 1, len(self.options) - 1))

        @kb.add("home", filter=~multiple)
        def _(event: KeyPressEvent) -> None:
            self.index = 0

        @kb.add("end", filter=~multiple)
        def _(event: KeyPressEvent) -> None:
            self.index = len(self.options) - 1

        @kb.add("left", filter=multiple)
        def _(event: KeyPressEvent) -> None:
            self.hovered = max(0, min((self.hovered or 0) - 1, len(self.options) - 1))
            get_app().layout.focus(self.buttons[self.hovered])

        @kb.add("right", filter=multiple)
        def _(event: KeyPressEvent) -> None:
            self.hovered = max(0, min((self.hovered or 0) + 1, len(self.options) - 1))
            get_app().layout.focus(self.buttons[self.hovered])

        @kb.add("home", filter=multiple)
        def _(event: KeyPressEvent) -> None:
            self.hovered = 0
            get_app().layout.focus(self.buttons[self.hovered])

        @kb.add("end", filter=multiple)
        def _(event: KeyPressEvent) -> None:
            self.hovered = len(self.options) - 1
            get_app().layout.focus(self.buttons[self.hovered])

        return kb


class SliderControl(UIControl):
    """A control to display a slider."""

    def __init__(
        self,
        slider: Slider,
        show_arrows: FilterOrBool = True,
        handle_char: str = "●",
        track_char: str | None = None,
        selected_track_char: str | None = None,
        style: str = "class:input",
        disabled: FilterOrBool = False,
    ) -> None:
        """Create a new slider control instance.

        Args:
            slider: The slider widget the control belongs to (provides slider data)
            show_arrows: Whether to show increment / decrement buttons at each end of
                the slider
            handle_char: The character to use as the slider handle
            track_char: The character to use for the slider track
            selected_track_char: The character to use for the selected section of the
                slider track
            style: A style string to apply to the slider
            disabled: A filter which when evaluated to :py:const:`True` causes the
                widget to be disabled
        """
        self.slider = slider

        self.track_char = track_char
        self.selected_track_char = selected_track_char
        self.show_arrows = to_filter(show_arrows)
        self.handle_char = handle_char
        self.style = style
        self.disabled = to_filter(disabled)

        self.selected_handle = 0
        self.track_len = 0

        self.mouse_handlers: dict[int, Callable[..., NotImplementedOrNone]] = {}
        self.dragging = False
        self.repeatable = True
        self.repeat_task: asyncio.Task[None] | None = None

        self._content_cache: SimpleCache = SimpleCache(maxsize=50)
        self.window: Window | None = None

    def preferred_width(self, max_available_width: int) -> int | None:
        """Return the preferred width of the slider control given its orientation."""
        return 1 if self.slider.vertical() else max_available_width

    def preferred_height(
        self,
        width: int,
        max_available_height: int,
        wrap_lines: bool,
        get_line_prefix: GetLinePrefixCallable | None,
    ) -> int | None:
        """Return the preferred height of the slider control given its orientation."""
        return min(max_available_height, 10) if self.slider.vertical() else 1

    def is_focusable(self) -> bool:
        """Tell whether this user control is focusable."""
        return not self.disabled()

    def create_content(self, width: int, height: int) -> UIContent:
        """Create an cache the rendered control fragments."""

        def get_content() -> UIContent:
            fragment_lines = self.render_lines(width, height)

            return UIContent(
                get_line=lambda i: fragment_lines[i],
                line_count=len(fragment_lines),
                show_cursor=False,
            )

        key = (
            width,
            self.slider.vertical(),
            tuple(self.slider.options),
            tuple(self.slider.mask),
            self.selected_handle,
            self.slider.has_focus(),
        )
        return self._content_cache.get(key, get_content)

    def set_index(
        self,
        handle: int = 0,
        ab: int | None = None,
        rel: int | None = None,
        fire: bool = True,
    ) -> NotImplementedOrNone:
        """Set the selected index of the slider."""
        assert ab is not None or rel is not None
        if rel is not None:
            ab = self.slider.indices[handle] + rel
        assert ab is not None
        # Restrict set value to permitted range
        ab = min(len(self.slider.options) - 1, max(0, ab))

        # Do not allow both handles to have the same value
        indices = dict(enumerate(self.slider.indices))
        if ab not in indices.values():
            # Set the handle's value
            indices.update({handle: ab})
            # Update the slider
            self.slider.indices = [index for handle, index in sorted(indices.items())]
            if fire:
                self.slider.on_change.fire()
            return None
        else:
            return NotImplemented

    @property
    def selected_handle(self) -> int:
        """Return the currently selected slider handle."""
        return self._selected_handle

    @selected_handle.setter
    def selected_handle(self, value: int) -> None:
        """Set the currently selected slider handle."""
        value = max(0, min(value, sum(self.slider.mask)))
        self._selected_handle = value

    def _draw_handle(self, n: int) -> OneStyleAndTextTuple:
        """Draw the given slider handle given it's current focus and selection state."""
        selected_style = "class:selection" if self.selected_handle == n else ""
        focused_style = "class:focused" if self.slider.has_focus() else ""
        return (
            f"class:handle {selected_style} {focused_style}",
            self.handle_char,
        )

    def mouse_handler_handle(
        self, mouse_event: MouseEvent, handle: int = 0
    ) -> NotImplementedOrNone:
        """Handle mouse events on the slider's handles."""
        if mouse_event.event_type == MouseEventType.MOUSE_DOWN:
            self.selected_handle = handle
            self.dragging = True

            # Set global mouse capture
            if (
                self.window is not None
                and (render_info := self.window.render_info) is not None
                and render_info._rowcol_to_yx
            ):
                y_min, x_min = min(render_info._rowcol_to_yx.values())
                y_max, x_max = max(render_info._rowcol_to_yx.values())
                get_app().mouse_limits = WritePosition(
                    xpos=x_min,
                    ypos=y_min,
                    width=x_max - x_min,
                    height=y_max - y_min,
                )
        return self.mouse_handler_scroll(mouse_event)

    def mouse_handler_track(
        self, mouse_event: MouseEvent, index: int = 0
    ) -> NotImplementedOrNone:
        """Handle mouse events on the slider track."""
        if mouse_event.event_type == MouseEventType.MOUSE_DOWN:
            get_app().layout.focus(self)
            handle = self.selected_handle
            if sum(self.slider.mask) == 2:
                if handle == 0 and index > self.slider.indices[1]:
                    self.selected_handle = 1
                elif handle == 1 and index < self.slider.indices[0]:
                    self.selected_handle = 0
            self.set_index(self.selected_handle, ab=index)
            return None
        else:
            if self.repeat_task is not None:
                self.repeat_task.cancel()
            return self.mouse_handler_scroll(mouse_event)

    def mouse_handler_arrow(
        self, mouse_event: MouseEvent, n: int = 0
    ) -> NotImplementedOrNone:
        """Handle mouse events on the slider's arrows."""
        if mouse_event.event_type == MouseEventType.MOUSE_DOWN:
            get_app().layout.focus(self)
            self.set_index(self.selected_handle, rel=n)
            # Trigger this mouse event to be repeated
            if self.repeatable:
                self.repeat_task = get_app().create_background_task(
                    self.repeat(mouse_event, handler=self.mouse_handler_arrow, n=n)
                )
            else:
                self.repeatable = True
            return None
        else:
            # Stop any repeated tasks
            self.repeatable = False
            if self.repeat_task is not None:
                self.repeat_task.cancel()
            return self.mouse_handler_scroll(mouse_event)

    def mouse_handler_scroll(
        self, mouse_event: MouseEvent, handle: int | None = None
    ) -> NotImplementedOrNone:
        """Handle mouse scroll events."""
        if self.slider.has_focus():
            if handle is None:
                handle = self.selected_handle
            if mouse_event.event_type == MouseEventType.SCROLL_UP:
                return self.set_index(handle, rel=1)
            elif mouse_event.event_type == MouseEventType.SCROLL_DOWN:
                return self.set_index(handle, rel=-1)
        return NotImplemented

    def mouse_handler_(self, mouse_event: MouseEvent, loc: int) -> NotImplementedOrNone:
        """Handle mouse events."""
        if self.dragging and mouse_event.event_type == MouseEventType.MOUSE_MOVE:
            n_options = len(self.slider.options)
            pos = loc
            if self.show_arrows():
                pos -= 2
            pos = max(0, min(self.track_len, pos))
            index = int((n_options - 0.5) * pos / self.track_len)
            if self.slider.vertical():
                index = n_options - index
            self.set_index(
                self.selected_handle,
                ab=index,
            )
            return None
        else:
            self.dragging = False
            get_app().mouse_limits = None
            # Call the underlying mouse handler
            if handler := self.mouse_handlers.get(loc):
                return handler(mouse_event)
            else:
                return NotImplemented

    async def repeat(
        self,
        mouse_event: MouseEvent,
        handler: Callable[..., NotImplementedOrNone],
        timeout: float = 0.25,
        **kwargs: Any,
    ) -> None:
        """Repeat a mouse event after a timeout."""
        await asyncio.sleep(timeout)
        handler(mouse_event, **kwargs)
        get_app().invalidate()

    def get_key_bindings(self) -> KeyBindingsBase | None:
        """Key bindings for the Slider."""
        kb = KeyBindings()

        if self.slider.vertical():

            @kb.add("down")
            def _(event: KeyPressEvent) -> None:
                self.set_index(self.selected_handle, rel=-1)

            @kb.add("up")
            def _(event: KeyPressEvent) -> None:
                self.set_index(self.selected_handle, rel=1)

        else:

            @kb.add("left")
            def _(event: KeyPressEvent) -> None:
                self.set_index(self.selected_handle, rel=-1)

            @kb.add("right")
            def _(event: KeyPressEvent) -> None:
                self.set_index(self.selected_handle, rel=1)

        @kb.add("home")
        def _(event: KeyPressEvent) -> None:
            self.set_index(self.selected_handle, ab=0)

        @kb.add("end")
        def _(event: KeyPressEvent) -> None:
            self.set_index(self.selected_handle, ab=len(self.slider.options))

        return kb

    def render_lines(self, width: int, height: int) -> list[StyleAndTextTuples]:
        """Generate formatted text fragments to display the slider."""
        ft = []
        mouse_handlers: list[Callable[..., NotImplementedOrNone]] = []

        vertical = self.slider.vertical()
        size = height if vertical else width
        n_handles = sum(self.slider.mask)
        n_options = len(self.slider.options)
        indices = self.slider.indices
        track_len = size - n_handles

        track_char = self.track_char or ("│" if vertical else "─")
        selected_track_char = self.selected_track_char or ("┃" if vertical else "━")

        if self.show_arrows():
            # The arrows take up 4 characters: remove them from the track length
            track_len -= 4
            ft += [
                *to_formatted_text(self.slider.arrows[0], "class:arrow,left"),
                ("", " "),
            ]
            mouse_handlers += [
                partial(self.mouse_handler_arrow, n=-1),
                self.mouse_handler_scroll,
            ]

        # First bit of track
        left_len = floor(track_len * indices[0] / (n_options - 1))
        ft.append(("class:track", track_char * left_len))
        mouse_handlers += [
            partial(
                self.mouse_handler_track,
                index=int((n_options - 0.5) * i / track_len),
            )
            for i in range(left_len)
        ]

        # First handle
        ft.append(self._draw_handle(0))
        mouse_handlers.append(partial(self.mouse_handler_handle, handle=0))

        # Middle bit of track
        middle_len = 0
        if n_handles > 1:
            middle_len = ceil(track_len * (indices[-1] - indices[0]) / (n_options - 1))
            ft.append(("class:track,selection", selected_track_char * middle_len))
            mouse_handlers += [
                partial(
                    self.mouse_handler_track,
                    index=int((n_options - 0.5) * i / track_len),
                )
                for i in range(left_len, left_len + middle_len)
            ]
            # Second handle
            ft.append(self._draw_handle(1))
            mouse_handlers.append(partial(self.mouse_handler_handle, handle=1))

        # Last bit of track
        right_len = track_len - left_len - middle_len
        ft.append(("class:track", track_char * right_len))
        mouse_handlers += [
            partial(
                self.mouse_handler_track,
                index=int(n_options * i / track_len),
            )
            for i in range(left_len + middle_len, track_len)
        ]

        if self.show_arrows():
            ft += [
                ("", " "),
                *to_formatted_text(self.slider.arrows[1], "class:arrow,right"),
            ]
            mouse_handlers += [
                self.mouse_handler_scroll,
                partial(self.mouse_handler_arrow, n=1),
            ]

        self.track_len = track_len

        if vertical:
            mouse_handlers = mouse_handlers[::-1]
            output = [[x] for x in explode_text_fragments(ft)][::-1]
        else:
            output = [ft]

        self.mouse_handlers = dict(enumerate(mouse_handlers))

        return output

    def mouse_handler(self, mouse_event: MouseEvent) -> NotImplementedOrNone:
        """Handle mouse events given the slider's orientation."""
        if self.slider.disabled():
            return None
        if self.slider.vertical():
            loc = mouse_event.position.y
        else:
            loc = mouse_event.position.x
        return self.mouse_handler_(mouse_event, loc=loc)


class Slider(SelectableWidget):
    """A slider widget with an optional editable readout."""

    control: SliderControl
    readout: Text

    def __init__(
        self,
        options: list[Any],
        labels: Sequence[AnyFormattedText] | None = None,
        index: int | None = None,
        indices: list[int] | None = None,
        n_values: int | None = None,
        multiple: FilterOrBool = False,
        max_count: int | None = None,
        on_change: Callable[[SelectableWidget], None] | None = None,
        style: str | Callable[[], str] = "class:input",
        border: GridStyle = InsetGrid,
        show_borders: DiBool | None = None,
        vertical: FilterOrBool = False,
        show_arrows: FilterOrBool = True,
        arrows: tuple[AnyFormattedText, AnyFormattedText] = ("-", "+"),
        show_readout: FilterOrBool = True,
        disabled: FilterOrBool = False,
    ) -> None:
        """Create a new slider widget instance.

        Args:
            options: A list of available options for the slider
            labels: An optional list of option names
            index: The index of the initially selected option if not a multiple
                selection slider
            indices: The indices of the start and end of the initially selected ranger
                of options if a range can be selected
            n_values: The number of values which are selectable
            multiple: If true, allow a range of options to be selected
            max_count: The maximum number of selectable items
            on_change: An optional function to call when the selection changes
            style: An optional style to apply to the widget
            border: The grid style to use for the borders (unused)
            show_borders: Whether the borders should be displayed (unused)
            vertical: If true, the slider will be displayed vertically
            show_arrows: If increment & decrement buttons should be shown at either end
                of the slider
            arrows: Strings to use for the increment and decrement buttons
            show_readout: If true, a read-out text box will be shown displaying the
                slider's current value
            disabled: A filter which when evaluated to :py:const:`True` causes the
                widget to be disabled

        """
        self.vertical = to_filter(vertical)
        self.arrows = arrows
        self.show_arrows = to_filter(show_arrows)
        self.show_readout = to_filter(show_readout)
        self.disabled = to_filter(disabled)

        if n_values is None and indices:
            n_values = len(indices)

        super().__init__(
            options=options,
            labels=labels,
            index=index,
            indices=indices,
            n_values=n_values,
            multiple=multiple,
            max_count=max_count,
            on_change=on_change,
            style=style,
            disabled=self.disabled,
        )
        self.on_change += self.value_changed

    def load_container(self) -> AnyContainer:
        """Build the slider's container."""
        self.control = SliderControl(
            slider=self, show_arrows=self.show_arrows, disabled=self.disabled
        )
        window = Window(self.control, style=lambda: f"class:slider {self.style}")
        self.control.window = window
        self.readout = Text(
            text=self.readout_text(self.indices),
            height=1,
            width=self.readout_len(),
            validation=lambda x: self.validate_readout(x) is not None,
            accept_handler=self.accept_handler,
            disabled=self.disabled,
            style=self.style,
        )
        return ConditionalSplit(
            self.vertical,
            [
                Box(window),
                ConditionalContainer(
                    Box(self.readout, padding=0),
                    filter=self.show_readout,
                ),
            ],
            padding=1,
            style=self._style,
        )

    def value_changed(self, slider: SelectableWidget | None = None) -> None:
        """Set the readout text when the slider value changes."""
        self.readout.text = self.readout_text(self.indices)

    def accept_handler(self, buffer: Buffer) -> bool:
        """Set the index to the value(s) entered in the readout buffer."""
        if values := self.validate_readout(buffer.text):
            for i, value in enumerate(values):
                self.control.set_index(
                    handle=i, ab=self.options.index(value), fire=False
                )
            # Trigger the event once all the values have been updated
            self.on_change.fire()
            return True
        return False

    def validate_readout(self, text: str) -> list[Any] | None:
        """Confirm the value entered in the readout is value."""
        values = [value.strip() for value in text.split("- ")]
        valid_values = []
        for value in values:
            for option in self.options:
                type_ = type(option)
                try:
                    typed_value = type_(value)
                except ValueError:
                    continue
                else:
                    if option == typed_value:
                        valid_values.append(typed_value)
                        break
            else:
                return None
        return valid_values

    def readout_text(self, indices: list[int]) -> str:
        """Return the readout text area value."""
        return " - ".join(map(str, (self.options[i] for i in indices)))

    def readout_len(self) -> int:
        """Return the length of the readout text area."""
        return min(
            10,
            len(
                self.readout_text(
                    [
                        x[1]
                        for x in sorted(
                            (len(str(x)), i) for i, x in enumerate(self.options)
                        )[-sum(self.mask) :]
                    ]
                )
            ),
        )



"""Defines a container which saves rendered output and re-displays it."""

from __future__ import annotations

import logging
from functools import cache
from typing import TYPE_CHECKING

from prompt_toolkit.cache import FastDictCache
from prompt_toolkit.data_structures import Point
from prompt_toolkit.layout.containers import (
    Container,
    Window,
    WindowRenderInfo,
    to_container,
)
from prompt_toolkit.layout.layout import walk
from prompt_toolkit.layout.mouse_handlers import MouseHandlers
from prompt_toolkit.mouse_events import MouseEvent

from euporie.core.app.current import get_app
from euporie.core.data_structures import DiInt
from euporie.core.layout.screen import BoundedWritePosition, Screen

if TYPE_CHECKING:
    from typing import Callable

    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.layout.dimension import Dimension
    from prompt_toolkit.layout.screen import Screen as PtkScreen
    from prompt_toolkit.layout.screen import WritePosition
    from prompt_toolkit.utils import Event

    MouseHandler = Callable[[MouseEvent], object]

log = logging.getLogger(__name__)


class CachedContainer(Container):
    """A container which renders its content once and caches the output."""

    def __init__(
        self,
        content: AnyContainer,
        mouse_handler_wrapper: Callable[[MouseHandler, CachedContainer], MouseHandler]
        | None = None,
    ) -> None:
        """Initiate the container."""
        self.content = content
        self.container = to_container(content)
        self.mouse_handler_wrapper = mouse_handler_wrapper

        self.screen = Screen()
        self.mouse_handlers = MouseHandlers()

        self._invalid = True
        self._invalidate_events: set[Event[object]] = set()
        self._layout_hash = 0
        self.render_counter = 0
        self.height = 0
        self.width = 0
        self._rendered_lines: set[int] = set()
        self._rowcols_to_yx: dict[Window, dict[tuple[int, int], tuple[int, int]]] = {}

        self._width_cache: FastDictCache[tuple[int, int], Dimension] = FastDictCache(
            get_value=lambda _render_count,
            max_available_width: self.container.preferred_width(max_available_width)
        )
        self._height_cache: FastDictCache[tuple[int, int, int], Dimension] = (
            FastDictCache(
                get_value=lambda _render_count,
                width,
                max_available_height: self.container.preferred_height(
                    width, max_available_height
                )
            )
        )

    @property
    def layout_hash(self) -> int:
        """Return a hash of the child's current layout."""
        return sum(hash(container) for container in walk(self.container))

    def invalidate(self) -> None:
        """Flag the child's rendering as out-of-date."""
        self._invalid = True

    def _invalidate_handler(self, sender: object) -> None:
        self.invalidate()

    def reset(self) -> None:
        """Reset the state of this container."""
        self.container.reset()
        self.invalidate()

    def preferred_width(self, max_available_width: int) -> Dimension:
        """Return the desired width for this container."""
        return self._width_cache[self.render_counter, max_available_width]

    def preferred_height(self, width: int, max_available_height: int) -> Dimension:
        """Return the desired height for this container."""
        return self._height_cache[self.render_counter, width, max_available_height]

    def render(
        self,
        available_width: int,
        available_height: int,
        style: str = "",
        start: int | None = None,
        end: int | None = None,
    ) -> None:
        """Render the child container at a given size.

        Args:
            available_width: The height available for rendering
            available_height: The width available for rendering
            style: The parent style to apply when rendering
            bbox: The bounding box for the content to render
            start: Rows between top of output and top of scrollable pane
            end: Rows between top of output and bottom of scrollable pane

        """
        if (
            self._layout_hash != (new_layout_hash := self.layout_hash)
            or self._invalid
            or self.width != available_width
        ):
            self._layout_hash = new_layout_hash
            self._rowcols_to_yx.clear()
            self._rendered_lines.clear()
            self.mouse_handlers.mouse_handlers.clear()
            self.screen = Screen()
            self.render_counter += 1

            # Recalculate child height if this child has been invalidated
            height = self.height = self.container.preferred_height(
                available_width, available_height
            ).preferred

        else:
            height = self.height

        # Calculate which lines are visible and which need rendering
        if start is not None:
            skip_top = max(0, -start)
        elif end is not None:
            skip_top = max(0, -end + height)
        else:
            skip_top = 0
        skip_bottom = max(0, height - available_height - skip_top)

        visible_lines = set(range(skip_top, height - skip_bottom))
        required_lines = visible_lines - self._rendered_lines

        # Refresh if needed
        if required_lines:
            screen = self.screen
            # TODO - allow horizontal scrolling too
            # self.width = self.container.preferred_width(available_width).width
            self.width = available_width

            self._invalid = False

            self.container.write_to_screen(
                screen,
                self.mouse_handlers,
                BoundedWritePosition(
                    xpos=0,
                    ypos=0,
                    width=self.width,
                    height=self.height,
                    # Only render lines not already on the current screen
                    bbox=DiInt(
                        top=min(required_lines),
                        right=0,
                        bottom=height - max(required_lines) - 1,
                        left=0,
                    ),
                ),
                style,
                erase_bg=True,
                z_index=0,
            )
            screen.draw_all_floats()

            events = set()
            rowcols_to_yx = self._rowcols_to_yx

            for container in walk(self.container):
                if isinstance(container, Window):
                    # Collect invalidation events
                    for event in container.content.get_invalidate_events():
                        event += self._invalidate_handler
                        events.add(event)

                    if (render_info := container.render_info) is not None:
                        # Update row/col to x/y mapping based on the lines we just rendereed
                        if container not in rowcols_to_yx:
                            rowcols_to_yx[container] = {}

                        rowcols_to_yx[container].update(render_info._rowcol_to_yx)
                        render_info._rowcol_to_yx = rowcols_to_yx[container]

            # Update the record of lines that've been rendered to the temporary screen
            self._rendered_lines |= required_lines

            # Remove handler from old invalidation events
            for event in events - set(self._invalidate_events):
                event -= self._invalidate_handler
            # Update the list of handlers
            self._invalidate_events = events

    def write_to_screen(
        self,
        screen: PtkScreen,
        mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        parent_style: str,
        erase_bg: bool,
        z_index: int | None,
    ) -> None:
        """Render the container to a `Screen` instance.

        Args:
            screen: The :class:`~prompt_toolkit.layout.screen.Screen` class to which
                the output has to be written.
            mouse_handlers: :class:`prompt_toolkit.layout.mouse_handlers.MouseHandlers`.
            write_position: A :class:`prompt_toolkit.layout.screen.WritePosition` object
                defining where this container should be drawn.
            erase_bg: If true, the background will be erased prior to drawing.
            parent_style: Style string to pass to the :class:`.Window` object. This will
                be applied to all content of the windows. :class:`.VSplit` and
                :class:`prompt_toolkit.layout.containers.HSplit` can use it to pass
                their style down to the windows that they contain.
            z_index: Used for propagating z_index from parent to child.

        """
        self.render(
            available_width=write_position.width,
            available_height=write_position.height,
            style=parent_style,
            # start: int | None = None,
            # end: int | None = None,
        )

        left = write_position.xpos
        top = write_position.ypos
        rows = slice(0, write_position.height)
        cols = slice(0, write_position.width)
        self.blit(screen, mouse_handlers, left, top, cols, rows)

    def blit(
        self,
        screen: PtkScreen,
        mouse_handlers: MouseHandlers,
        left: int,
        top: int,
        cols: slice,
        rows: slice,
    ) -> None:
        """Copy the rendered child from the local screen to the main screen.

        All locations are adjusted, allowing the pre-rendered child to be placed at any
        location on the main screen.

        Args:
            screen: The main screen to copy the pre-rendered screen data to
            mouse_handlers: The mouse handler collection to copy the pre-rendered
                handelers to
            left: The left-most column in which to start placing the data
            top: The upper row in which to start placing the data
            cols: The columns to copy
            rows: The rows to copy
                                                                                                                                                                                       .
        """
        # Copy write positions
        new_wps = {}

        for win, wp in self.screen.visible_windows_to_write_positions.items():
            new_wp = BoundedWritePosition(
                xpos=wp.xpos + left,
                ypos=wp.ypos + top,
                width=wp.width,
                height=wp.height,
                bbox=DiInt(
                    top=max(0, rows.start - wp.ypos),
                    right=max(0, wp.width - cols.stop),
                    bottom=max(0, wp.height - (rows.stop - wp.ypos)),
                    left=max(0, wp.width - (cols.stop - wp.xpos)),
                ),
            )
            new_wps[win] = new_wp

            # Modify render info
            info = win.render_info
            if info is not None:
                xpos = new_wp.xpos + info._x_offset
                y_start = new_wp.ypos + info._y_offset
                # Only include visible lines according to the bounding box
                visible_line_to_row_col = {
                    i: (y, xpos)
                    for i, y in enumerate(
                        range(
                            y_start + new_wp.bbox.top,
                            y_start + new_wp.height - new_wp.bbox.bottom,
                        )
                    )
                }
                win.render_info = WindowRenderInfo(
                    window=win,
                    ui_content=info.ui_content,
                    horizontal_scroll=0,
                    vertical_scroll=info.vertical_scroll,
                    window_width=info.window_width,
                    window_height=info.window_height,
                    configured_scroll_offsets=info.configured_scroll_offsets,
                    visible_line_to_row_col=visible_line_to_row_col,
                    # The following is needed to calculate absolute cursor positions
                    rowcol_to_yx={
                        k: (y + top, x + left)
                        for k, (y, x) in info._rowcol_to_yx.items()
                    },
                    x_offset=info._x_offset + left,
                    y_offset=info._y_offset + top,
                    wrap_lines=info.wrap_lines,
                )
                # Set horizontal scroll offset - TODO - fix this upstream
                if (
                    horizontal_scroll := getattr(info, "horizontal_scroll", None)
                ) is not None:
                    setattr(  # noqa B010
                        win.render_info, "horizontal_scroll", horizontal_scroll
                    )

        screen.visible_windows_to_write_positions.update(new_wps)
        screen.height = max(screen.height, self.screen.height)

        @cache
        def _wrap_mouse_handler(handler: Callable) -> MouseHandler:
            def _wrapped(mouse_event: MouseEvent) -> NotImplementedOrNone:
                # Modify mouse events to reflect position of content
                new_event = MouseEvent(
                    position=Point(
                        x=mouse_event.position.x - left,
                        y=mouse_event.position.y - top,
                    ),
                    event_type=mouse_event.event_type,
                    button=mouse_event.button,
                    modifiers=mouse_event.modifiers,
                )
                if callable(wrapper := self.mouse_handler_wrapper):
                    return wrapper(handler, self)(new_event)
                return handler(new_event)

            return _wrapped

        # Copy screen contents
        input_db = self.screen.data_buffer
        input_zwes = self.screen.zero_width_escapes
        input_mhs = self.mouse_handlers.mouse_handlers
        output_db = screen.data_buffer
        output_zwes = screen.zero_width_escapes
        output_mhs = mouse_handlers.mouse_handlers

        rows_range = range(max(0, rows.start), rows.stop)
        cols_range = range(max(0, cols.start), cols.stop)

        for y in rows_range:
            input_db_row = input_db[y]
            input_zwes_row = input_zwes[y]
            input_mhs_row = input_mhs[y]
            output_dbs_row = output_db[top + y]
            output_zwes_row = output_zwes[top + y]
            output_mhs_row = output_mhs[top + y]
            for x in cols_range:
                # Data
                output_dbs_row[left + x] = input_db_row[x]
                # Escape sequences
                output_zwes_row[left + x] = input_zwes_row[x]
                # Mouse handlers
                output_mhs_row[left + x] = _wrap_mouse_handler(input_mhs_row[x])

        # Copy cursors
        layout = get_app().layout
        if self.screen.show_cursor:
            for window, point in self.screen.cursor_positions.items():
                if (
                    layout.current_control == window.content
                    and window.render_info is not None
                    and window.render_info.ui_content.show_cursor
                    and not window.always_hide_cursor()
                    and cols.start <= point.x < cols.stop
                    and rows.start <= point.y < rows.stop
                ):
                    screen.cursor_positions[window] = Point(
                        x=left + point.x, y=top + point.y
                    )
                    screen.show_cursor = True

        # Copy menu positions
        screen.menu_positions.update(
            {
                window: Point(x=left + point.x, y=top + point.y)
                for window, point in self.screen.menu_positions.items()
            }
        )

    def get_children(self) -> list[Container]:
        """Return a list of all child containers."""
        return [self.container]



"""Define a cell object with input are and rich outputs, and related objects."""

from __future__ import annotations

import logging
from functools import lru_cache
from typing import TYPE_CHECKING

from prompt_toolkit.auto_suggest import AutoSuggest, DynamicAutoSuggest
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.completion.base import Completer, DynamicCompleter
from prompt_toolkit.document import Document
from prompt_toolkit.filters import (
    Condition,
    buffer_has_focus,
    has_focus,
    has_selection,
    is_done,
    is_searching,
    is_true,
    to_filter,
)
from prompt_toolkit.key_binding.key_bindings import merge_key_bindings
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    ScrollOffsets,
)
from prompt_toolkit.layout.controls import (
    BufferControl,
    GetLinePrefixCallable,
)
from prompt_toolkit.layout.dimension import AnyDimension
from prompt_toolkit.layout.dimension import Dimension as D
from prompt_toolkit.layout.margins import ConditionalMargin
from prompt_toolkit.layout.processors import (  # HighlightSearchProcessor,
    BeforeInput,
    ConditionalProcessor,
    DisplayMultipleCursors,
    HighlightIncrementalSearchProcessor,
    HighlightMatchingBracketProcessor,
    HighlightSelectionProcessor,
    PasswordProcessor,
    Processor,
    TabsProcessor,
)
from prompt_toolkit.lexers import DynamicLexer, PygmentsLexer, SimpleLexer
from prompt_toolkit.validation import DynamicValidator, Validator
from prompt_toolkit.widgets import TextArea
from prompt_toolkit.widgets.toolbars import SearchToolbar
from pygments.lexers import ClassNotFound, get_lexer_by_name

from euporie.core.app.current import get_app
from euporie.core.commands import add_cmd
from euporie.core.diagnostics import Report
from euporie.core.filters import buffer_is_code, scrollable
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)
from euporie.core.layout.containers import VSplit, Window
from euporie.core.margins import (
    MarginContainer,
    NumberedMargin,
    OverflowMargin,
    ScrollbarMargin,
)
from euporie.core.processors import (
    AppendLineAutoSuggestion,
    DiagnosticProcessor,
    ShowTrailingWhiteSpaceProcessor,
)
from euporie.core.suggest import ConditionalAutoSuggestAsync

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Callable

    from prompt_toolkit.buffer import BufferAcceptHandler
    from prompt_toolkit.filters import Filter, FilterOrBool
    from prompt_toolkit.formatted_text import (
        AnyFormattedText,
    )
    from prompt_toolkit.formatted_text.base import StyleAndTextTuples
    from prompt_toolkit.history import History
    from prompt_toolkit.key_binding.key_bindings import KeyBindingsBase
    from prompt_toolkit.layout.containers import AnyContainer, Container
    from prompt_toolkit.layout.layout import FocusableElement
    from prompt_toolkit.layout.margins import Margin
    from prompt_toolkit.lexers.base import Lexer

    from euporie.core.bars.status import StatusBarFields
    from euporie.core.format import Formatter
    from euporie.core.inspection import Inspector
    from euporie.core.tabs.kernel import KernelTab


log = logging.getLogger(__name__)


@lru_cache
def _get_lexer(highlight: bool, lexer: Lexer | None, language: str) -> Lexer:
    """Determine which lexer should be used for syntax highlighting."""
    if not highlight:
        return SimpleLexer()
    elif lexer is not None:
        return lexer
    try:
        pygments_lexer_class = get_lexer_by_name(language).__class__
    except ClassNotFound:
        return SimpleLexer()
    else:
        return PygmentsLexer(pygments_lexer_class, sync_from_start=False)


class KernelInput(TextArea):
    """Kernel input text areas.

    A customized text area for the cell input.
    """

    def __init__(
        self,
        kernel_tab: KernelTab,
        text: str = "",
        multiline: FilterOrBool = True,
        password: FilterOrBool = False,
        lexer: Lexer | None = None,
        auto_suggest: AutoSuggest | None = None,
        completer: Completer | None = None,
        complete_while_typing: FilterOrBool = True,
        validator: Validator | None = None,
        accept_handler: BufferAcceptHandler | None = None,
        history: History | None = None,
        focusable: FilterOrBool = True,
        focus_on_click: FilterOrBool = True,
        wrap_lines: FilterOrBool = False,
        read_only: FilterOrBool = False,
        width: AnyDimension = None,
        height: AnyDimension = None,
        dont_extend_height: FilterOrBool = False,
        dont_extend_width: FilterOrBool = False,
        line_numbers: bool = False,
        get_line_prefix: GetLinePrefixCallable | None = None,
        scrollbar: FilterOrBool = True,
        style: str = "class:kernel-input",
        search_field: SearchToolbar | None = None,
        preview_search: FilterOrBool = False,
        prompt: AnyFormattedText = "",
        input_processors: list[Processor] | None = None,
        name: str = "",
        left_margins: Sequence[Margin] | None = None,
        right_margins: Sequence[Margin] | None = None,
        on_text_changed: Callable[[Buffer], None] | None = None,
        on_cursor_position_changed: Callable[[Buffer], None] | None = None,
        tempfile_suffix: str | Callable[[], str] = "",
        key_bindings: KeyBindingsBase | None = None,
        enable_history_search: FilterOrBool = False,
        autosuggest_while_typing: FilterOrBool = True,
        validate_while_typing: FilterOrBool = False,
        scroll_offsets: ScrollOffsets | None = None,
        formatters: list[Formatter] | None = None,
        language: str | Callable[[], str] | None = None,
        diagnostics: Report | Callable[[], Report] | None = None,
        inspector: Inspector | None = None,
        show_diagnostics: FilterOrBool = True,
    ) -> None:
        """Initiate the cell input box."""
        self.kernel_tab = kernel_tab
        app = kernel_tab.app

        if history is None:
            history = kernel_tab.history

        if search_field is None:
            search_field = app.search_bar
        if isinstance(search_field, SearchToolbar):
            search_control = search_field.control
        else:
            search_control = None

        if input_processors is None:
            input_processors = []

        # Writeable attributes.
        self.completer = completer
        self.complete_while_typing = Condition(
            lambda: app.config.autocomplete
        ) & to_filter(complete_while_typing)
        self.auto_suggest = auto_suggest or ConditionalAutoSuggestAsync(
            self.kernel_tab.suggester,
            filter=to_filter(autosuggest_while_typing)
            & Condition(lambda: app.config.autosuggest),
        )
        self.read_only = read_only
        self.wrap_lines = wrap_lines
        self.validator = validator
        self._language = language
        self.lexer = lexer

        self.formatters = formatters if formatters is not None else []
        self._diagnostics = diagnostics or Report()
        self.inspector = inspector

        self.buffer = Buffer(
            document=Document(text, 0),
            multiline=multiline,
            read_only=Condition(lambda: is_true(self.read_only)),
            completer=DynamicCompleter(lambda: self.completer),
            complete_while_typing=Condition(
                lambda: is_true(self.complete_while_typing)
            ),
            validator=DynamicValidator(lambda: self.validator),
            auto_suggest=DynamicAutoSuggest(lambda: self.auto_suggest),
            accept_handler=accept_handler,
            history=history,
            name=name,
            validate_while_typing=validate_while_typing,
            on_text_changed=on_text_changed,
            on_cursor_position_changed=on_cursor_position_changed,
            tempfile_suffix=tempfile_suffix or kernel_tab.kernel_lang_file_ext,
            enable_history_search=to_filter(enable_history_search),
        )
        self.has_focus: Filter = has_focus(self.buffer)

        # Set up autoinspect
        def _on_cursor_position_changed(buf: Buffer) -> None:
            """Respond to cursor movements."""
            # Update contextual help
            if app.config.autoinspect and self.buffer.name == "code":
                app.create_background_task(self.inspect(auto=True))
            elif pager := app.pager:
                pager.hide()

        self.buffer.on_cursor_position_changed += _on_cursor_position_changed

        # Set extra key-bindings
        widgets_key_bindings = load_registered_bindings(
            "euporie.core.widgets.inputs.KernelInput",
            config=app.config,
        )
        if key_bindings:
            widgets_key_bindings = merge_key_bindings(
                [key_bindings, widgets_key_bindings]
            )

        def _get_diagnostics() -> Report:
            return self.diagnostics

        self.control = BufferControl(
            buffer=self.buffer,
            lexer=DynamicLexer(
                lambda: _get_lexer(
                    # Only lex buffers with text
                    self.buffer.text and app.config.syntax_highlighting,
                    self.lexer,
                    self.language,
                )
            ),
            input_processors=[
                ConditionalProcessor(
                    DiagnosticProcessor(_get_diagnostics),
                    filter=to_filter(show_diagnostics),
                ),
                ConditionalProcessor(  # type: ignore
                    AppendLineAutoSuggestion(),
                    has_focus(self.buffer) & ~is_done,
                ),
                ConditionalProcessor(
                    processor=PasswordProcessor(), filter=to_filter(password)
                ),
                BeforeInput(prompt, style="class:text-area.prompt"),
                ConditionalProcessor(
                    HighlightIncrementalSearchProcessor(),
                    filter=is_searching,
                ),
                # HighlightSearchProcessor(),
                HighlightSelectionProcessor(),
                DisplayMultipleCursors(),
                ConditionalProcessor(  # type: ignore
                    HighlightMatchingBracketProcessor(),
                    has_focus(self.buffer) & ~is_done,
                ),
                TabsProcessor(char1="⇥", char2="┈"),
                ShowTrailingWhiteSpaceProcessor(),
                *input_processors,
            ],
            search_buffer_control=search_control,
            preview_search=preview_search,
            focusable=focusable,
            focus_on_click=focus_on_click,
            include_default_input_processors=False,
            key_bindings=widgets_key_bindings,
        )

        left_margins = [
            ConditionalMargin(
                NumberedMargin(
                    diagnostics=_get_diagnostics,
                    show_diagnostics=to_filter(show_diagnostics),
                ),
                app.config.filters.line_numbers & self.buffer.multiline,
            ),
        ]
        right_margins = [OverflowMargin()]
        self.window = Window(
            height=lambda: height or D(min=1)
            if self.buffer.multiline()
            else D.exact(1),
            width=width,
            dont_extend_height=dont_extend_height,
            dont_extend_width=dont_extend_width,
            content=self.control,
            wrap_lines=Condition(lambda: is_true(self.wrap_lines)),
            left_margins=left_margins,
            right_margins=right_margins,
            get_line_prefix=get_line_prefix,
            style=lambda: (
                "class:text-area "
                + ("class:focused " if self.has_focus() else "")
                + style
            ),
            cursorline=has_focus(self.buffer),
            scroll_offsets=scroll_offsets
            or ScrollOffsets(top=1, right=1, bottom=1, left=1),
        )

        self.container = VSplit(
            [
                self.window,
                ConditionalContainer(
                    MarginContainer(ScrollbarMargin(), target=self.window),
                    filter=to_filter(scrollbar) & scrollable(self.window),
                ),
            ]
        )

    @property
    def language(self) -> str:
        """The current language of the text in the input box."""
        return str(
            self._language() if callable(self._language) else self.language
        ).casefold()

    @language.setter
    def language(self, value: str) -> None:
        """Set the current language of the text in the input box."""
        self._language = value

    @property
    def diagnostics(self) -> Report:
        """The current diagnostics report."""
        if callable(self._diagnostics):
            return self._diagnostics()
        return self._diagnostics

    @diagnostics.setter
    def diagnostics(self, value: Report) -> None:
        """Set the current diagnostics report."""
        self._diagnostics = value

    def reformat(self) -> None:
        """Reformat the cell's input."""
        text = self.buffer.text
        language = self.language
        for formatter in self.formatters:
            text = formatter._format(text, language)
        self.buffer.text = text

    async def inspect(self, auto: bool = False) -> None:
        """Get contextual help for the current cursor position in the current cell."""
        from euporie.core.widgets.pager import PagerState

        if self.inspector is None:
            return

        document = self.buffer.document
        pager = self.kernel_tab.app.pager
        assert pager is not None

        prev_state = pager.state
        if (
            not auto
            and pager.visible()
            and prev_state is not None
            and prev_state.code == document.text
            and prev_state.cursor_pos == document.cursor_position
        ):
            pager.focus()
            return

        data = await self.inspector.get_context(document, auto=auto)
        new_state = (
            PagerState(
                code=document.text, cursor_pos=document.cursor_position, data=data
            )
            if data
            else None
        )
        if prev_state != new_state:
            pager.state = new_state
            self.kernel_tab.app.invalidate()

    @property
    def current_diagnostic_message(self) -> StyleAndTextTuples:
        """Format the currently selected diagnostic message."""
        document = self.buffer.document
        row = document.cursor_position_row
        col = document.cursor_position_col
        line: StyleAndTextTuples = []
        # mypy bug https://github.com/python/mypy/issues/16733
        for diagnostic in self.diagnostics:  # type: ignore [attr-defined]
            lines = diagnostic.lines
            chars = diagnostic.chars
            if lines.start <= row < lines.stop and chars.start <= col < chars.stop:
                if diagnostic.code:
                    line.append(
                        (f"class:diagnostic-{diagnostic.level}", diagnostic.code)
                    )
                if diagnostic.message:
                    if line:
                        line.append(("", " - "))
                    line.append(("", diagnostic.message))
                    break
        return line

    def __pt_status__(self) -> StatusBarFields | None:
        """Return a list of statusbar field values shown then this tab is active."""
        return ([self.current_diagnostic_message], [])

    def __pt_container__(self) -> Container:
        """Return the widget's container."""
        return self.container

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd(
        filter=buffer_is_code & buffer_has_focus & ~has_selection,
    )
    async def _show_contextual_help() -> None:
        """Display contextual help."""
        from euporie.core.tabs.kernel import KernelTab

        tab = get_app().tab
        if isinstance(tab, KernelTab) and (input_box := tab.current_input) is not None:
            await input_box.inspect()

    @staticmethod
    @add_cmd(filter=buffer_is_code & buffer_has_focus)
    def _history_prev() -> None:
        """Get the previous history entry."""
        from euporie.core.app.current import get_app

        get_app().current_buffer.history_backward()

    @staticmethod
    @add_cmd(filter=buffer_is_code & buffer_has_focus)
    def _history_next() -> None:
        """Get the next history entry."""
        from euporie.core.app.current import get_app

        get_app().current_buffer.history_forward()

    @staticmethod
    @add_cmd()
    def _reformat_input() -> None:
        """Format the contents of the current input field."""
        from euporie.core.app.current import get_app
        from euporie.core.tabs.kernel import KernelTab

        if (
            (tab := get_app().tab)
            and (isinstance(tab, KernelTab))
            and (current_input := tab.current_input)
        ):
            current_input.reformat()

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.core.widgets.inputs.KernelInput": {
                "show-contextual-help": "s-tab",
                "history-prev": "c-A-up",
                "history-next": "c-A-down",
            }
        }
    )


class StdInput:
    """A widget to accept kernel input."""

    def __init__(self, kernel_tab: KernelTab) -> None:
        """Create a new kernel input box."""
        from euporie.core.widgets.forms import LabelledWidget, Text

        self.kernel_tab = kernel_tab
        self.last_focused: FocusableElement | None = None

        self.password = False
        self.prompt: str | None = None
        self.active = False
        self.visible = Condition(lambda: self.active)

        text = Text(
            multiline=False,
            accept_handler=self.accept,
            password=Condition(lambda: self.password),
            style="class:input",
        )
        self.window = text.window
        self.container = ConditionalContainer(
            LabelledWidget(
                body=text,
                label=lambda: self.prompt or ">>>",
            ),
            filter=self.visible,
        )

    def accept(self, buffer: Buffer) -> bool:
        """Send the input to the kernel and hide the input box."""
        if self.kernel_tab.kernel is not None:
            self.kernel_tab.kernel.input(buffer.text)
        # Cleanup
        self.active = False

        buffer.text = ""
        if self.last_focused:
            try:
                get_app().layout.focus(self.last_focused)
            except ValueError:
                pass
        return True

    def get_input(
        self,
        prompt: str = "Please enter a value: ",
        password: bool = False,
    ) -> None:
        """Prompt the user for input and sends the result to the kernel."""
        self.password = password
        self.prompt = prompt

        # Set this first so the height of the cell includes the input box if it gets
        # rendered when we scroll to it
        self.active = True
        # Remember what was focused before
        app = get_app()
        layout = app.layout
        self.last_focused = layout.current_control

        # Try focusing the input box - we create an asynchronous task which will
        # probably run after the next render, when the stdin_box is recognised as being
        # in the layout. This doesn't always work (depending on timing), does usually.
        try:
            layout.focus(self)
        finally:

            async def _focus_input() -> None:
                # Focus the input box
                if self.window in layout.visible_windows:
                    layout.focus(self)
                # Redraw the screen to show it as focused
                app.invalidate()

            app.create_background_task(_focus_input())

    def __pt_container__(self) -> AnyContainer:
        """Return the input's container."""
        return self.container



"""Contain lexers for pygments."""

from __future__ import annotations

from typing import ClassVar

from pygments.lexer import RegexLexer
from pygments.style import Style
from pygments.token import (
    Comment,
    Error,
    Generic,
    Keyword,
    Literal,
    Name,
    Operator,
    Text,
    _TokenType,
)


class ArgparseLexer(RegexLexer):
    """A pygments lexer for agrparse help text."""

    name = "argparse"
    aliases: ClassVar[list[str]] = ["argparse"]
    filenames: ClassVar[list[str]] = []

    tokens: ClassVar[
        dict[str, list[tuple[str, _TokenType] | tuple[str, _TokenType, str]]]
    ] = {
        "root": [
            (r"(?<=usage: )[^\s]+", Name.Namespace),
            (r"\{", Operator, "options"),
            (r"[\[\{\|\}\]]", Operator),
            (r"((?<=\s)|(?<=\[))(--[a-zA-Z0-9-]+|-[a-zA-Z0-9-])", Keyword),
            (r"^(\w+\s)?\w+:", Generic.Heading),
            (r"\b(str|int|bool|UPath|loads)\b", Name.Builtin),
            (r"\b[A-Z]+_[A-Z]*\b", Name.Variable),
            (r"'.*?'", Literal.String),
            (r".", Text),
        ],
        "options": [
            (r"\d+", Literal.Number),
            (r",", Text),
            (r"[^\}]", Literal.String),
            (r"\}", Operator, "#pop"),
        ],
    }


class EuporiePygmentsStyle(Style):
    """Version of pygment's "native" style which works better on light backgrounds."""

    styles: ClassVar[dict[_TokenType, str]] = {
        Comment: "italic #888888",
        Comment.Preproc: "noitalic bold #cd2828",
        Comment.Special: "noitalic bold #e50808 bg:#520000",
        Keyword: "bold #6ebf26",
        Keyword.Pseudo: "nobold",
        Keyword.Constant: "nobold #ff3d3d",
        Operator.Word: "bold #6ebf26",
        Literal.Date: "#2fbccd",
        Literal.String: "#ed9d13",
        Literal.String.Other: "#ffa500",
        Literal.Number: "#51b2fd",
        Name.Builtin: "#2fbccd",
        Name.Variable: "#40ffff",
        Name.Constant: "#40ffff",
        Name.Class: "underline #71adff",
        Name.Function: "#71adff",
        Name.Namespace: "underline #71adff",
        Name.Exception: "noinherit bold",
        Name.Tag: "bold #6ebf26",
        Name.Attribute: "noinherit",
        Name.Decorator: "#ffa500",
        Generic.Heading: "bold",
        Generic.Subheading: "underline",
        Generic.Deleted: "#d22323",
        Generic.Inserted: "#589819",
        Generic.Error: "#d22323",
        Generic.Emph: "italic",
        Generic.Strong: "bold",
        Generic.Traceback: "#d22323",
        Error: "bold bg:#a61717 #ffffff",
    }



"""Defines a container which displays all children at full height vertially stacked."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from prompt_toolkit.application.current import get_app
from prompt_toolkit.layout.containers import (
    Container,
    Window,
    to_container,
)
from prompt_toolkit.layout.dimension import Dimension, to_dimension

from euporie.core.layout.screen import BoundedWritePosition

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Callable

    from prompt_toolkit.key_binding.key_bindings import (
        KeyBindingsBase,
    )
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.layout.dimension import AnyDimension
    from prompt_toolkit.layout.mouse_handlers import MouseHandlers
    from prompt_toolkit.layout.screen import Screen, WritePosition
    from prompt_toolkit.mouse_events import MouseEvent

    MouseHandler = Callable[[MouseEvent], object]

log = logging.getLogger(__name__)


class PrintingContainer(Container):
    """A container which displays all it's children in a vertical list."""

    def __init__(
        self,
        children: Callable | Sequence[AnyContainer],
        width: AnyDimension = None,
        key_bindings: KeyBindingsBase | None = None,
    ) -> None:
        """Initiate the container."""
        self.width = width
        self._children = children
        self._render_count = -1
        self._cached_children: Sequence[AnyContainer] | None = None
        self.key_bindings = key_bindings

    def get_key_bindings(self) -> KeyBindingsBase | None:
        """Return the container's key bindings."""
        return self.key_bindings

    @property
    def children(self) -> Sequence[AnyContainer]:
        """Return the container's children."""
        # Only load the children from a callable once per render cycle
        if callable(self._children):
            if (
                self._cached_children is None
                or self._render_count != get_app().render_counter
            ):
                self._cached_children = self._children()
                self._render_count = get_app().render_counter
            children = self._cached_children
        else:
            children = self._children
        return children or [Window()]

    def get_children(self) -> list[Container]:
        """Return a list of all child containers."""
        return list(map(to_container, self.children))

    def write_to_screen(
        self,
        screen: Screen,
        mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        parent_style: str,
        erase_bg: bool,
        z_index: int | None,
    ) -> None:
        """Render the container to a `Screen` instance.

        All children are rendered vertically in sequence.

        Args:
            screen: The :class:`~prompt_toolkit.layout.screen.Screen` class to which
                the output has to be written.
            mouse_handlers: :class:`prompt_toolkit.layout.mouse_handlers.MouseHandlers`.
            write_position: A :class:`prompt_toolkit.layout.screen.WritePosition` object
                defining where this container should be drawn.
            erase_bg: If true, the background will be erased prior to drawing.
            parent_style: Style string to pass to the :class:`.Window` object. This will
                be applied to all content of the windows. :class:`.VSplit` and
                :class:`prompt_toolkit.layout.containers.HSplit` can use it to pass
                their style down to the windows that they contain.
            z_index: Used for propagating z_index from parent to child.

        """
        xpos = write_position.xpos
        ypos = write_position.ypos

        children = self.get_children()
        for child in children:
            height = child.preferred_height(write_position.width, 999999).preferred
            child.write_to_screen(
                screen,
                mouse_handlers,
                BoundedWritePosition(xpos, ypos, write_position.width, height),
                parent_style,
                erase_bg,
                z_index,
            )
            ypos += height

    def preferred_height(self, width: int, max_available_height: int) -> Dimension:
        """Return the preferred height, equal to the sum of the child heights."""
        return Dimension(
            min=1,
            preferred=sum(
                [
                    c.preferred_height(width, max_available_height).preferred
                    for c in self.get_children()
                ]
            ),
        )

    def preferred_width(self, max_available_width: int) -> Dimension:
        """Calculate and returns the desired width for this container."""
        if self.width is not None:
            dim = to_dimension(self.width).preferred
            return Dimension(max=dim, preferred=dim)
        else:
            return Dimension(max_available_width)

    def reset(self) -> None:
        """Reset the state of this container and all the children.

        Does nothing as this container is used for dumping output.
        """



"""Contain main format conversion function."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, NamedTuple

from prompt_toolkit.cache import FastDictCache, SimpleCache
from prompt_toolkit.filters import to_filter

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Callable

    from prompt_toolkit.filters import Filter, FilterOrBool

log = logging.getLogger(__name__)


class Converter(NamedTuple):
    """Hold a conversion function and its weight."""

    func: Callable
    filter_: Filter
    weight: int = 1


converters: dict[str, dict[str, list[Converter]]] = {}

_FILTER_CACHE: SimpleCache = SimpleCache()


def register(
    from_: Iterable[str] | str,
    to: str,
    filter_: FilterOrBool = True,
    weight: int = 1,
) -> Callable:
    """Add a converter to the centralized format conversion system."""
    if isinstance(from_, str):
        from_ = (from_,)

    def decorator(func: Callable) -> Callable:
        if to not in converters:
            converters[to] = {}
        for from_format in from_:
            if from_format not in converters[to]:
                converters[to][from_format] = []
            converters[to][from_format].append(
                Converter(func=func, filter_=to_filter(filter_), weight=weight)
            )
        return func

    return decorator


def _find_route(from_: str, to: str) -> list | None:
    """Find the shortest conversion path between two formats."""
    from euporie.core.convert import formats  # noqa: F401

    if from_ == to:
        return [from_]

    chains = []

    def find(start: str, chain: list[str]) -> None:
        if chain[0] == start:
            chains.append(chain)
        sources: dict[str, list[Converter]] = converters.get(chain[0], {})
        for link in sources:
            if link not in chain and any(
                _FILTER_CACHE.get((conv,), conv.filter_)
                for conv in sources.get(link, [])
            ):
                find(start, [link, *chain])

    find(from_, [to])

    if chains:
        # Find chain with shortest weighted length
        return sorted(
            chains,
            key=lambda chain: sum(
                [
                    min(
                        [
                            conv.weight
                            for conv in converters.get(step_b, {}).get(step_a, [])
                            if _FILTER_CACHE.get((conv,), conv.filter_)
                        ]
                    )
                    for step_a, step_b in zip(chain, chain[1:])
                ]
            ),
        )
    else:
        return None


_CONVERTOR_ROUTE_CACHE: FastDictCache[tuple[str, str], list | None] = FastDictCache(
    _find_route
)


def find_route(from_: str, to: str) -> list | None:
    """Find and cache conversion routes."""
    return _CONVERTOR_ROUTE_CACHE[from_, to]



"""Defines a logo widget."""

from __future__ import annotations

from typing import TYPE_CHECKING, Callable

from prompt_toolkit.layout.containers import WindowAlign
from prompt_toolkit.layout.controls import FormattedTextControl

from euporie.core import __logo__
from euporie.core.filters import has_tabs
from euporie.core.layout.containers import Window

if TYPE_CHECKING:
    from typing import Any

    from prompt_toolkit.filters import FilterOrBool
    from prompt_toolkit.layout.controls import UIControl
    from prompt_toolkit.layout.dimension import AnyDimension


class Logo(Window):
    """A widget to display the application's logo."""

    def __init__(
        self,
        content: UIControl | None = None,
        height: AnyDimension = 1,
        width: AnyDimension = 3,
        style: str | Callable[[], str] = "class:menu,logo",
        dont_extend_width: FilterOrBool = True,
        align: WindowAlign | Callable[[], WindowAlign] = WindowAlign.CENTER,
        **kwargs: Any,
    ) -> None:
        """Create a new window with defaults specified."""
        if content is None:
            content = FormattedTextControl(
                [("", f" {__logo__} ")],
                focusable=~has_tabs,
                show_cursor=False,
            )
        super().__init__(
            content=content,
            height=height,
            width=width,
            style=style,
            dont_extend_width=dont_extend_width,
            **kwargs,
        )



"""Utility functions for format converters."""

from __future__ import annotations

import asyncio
import logging
import subprocess  # S404 - Security implications have been considered
import tempfile
from math import ceil
from pathlib import Path
from typing import TYPE_CHECKING

from euporie.core.app.current import get_app

if TYPE_CHECKING:
    from typing import Any

    from euporie.core.convert.datum import Datum

log = logging.getLogger(__name__)


async def scale_to_fit(
    datum: Datum, cols: int | None, rows: int | None
) -> tuple[int, int]:
    """Calculate image size based on aspect ratio, and scale to fit."""
    data = datum.data
    px, py = get_app().cell_size_px

    # Calculate rows based on image aspect ratio
    w, h = data.size
    if rows is None and cols is not None:
        rows = ceil(cols / w * h)
    elif cols is None and rows is not None:
        cols = ceil(rows / h * w)
    elif rows is None and cols is None:
        cols = ceil(w / px)
        rows = ceil(h / py)
    assert rows is not None
    assert cols is not None

    # Scale to fit while maintaining aspect ratio
    _width, aspect = await datum.cell_size_async()
    if cols * aspect < rows:
        rows = ceil(cols * aspect)
    else:
        cols = ceil(rows / aspect)

    return cols, rows


async def call_subproc(
    data: str | bytes,
    cmd: list[Any],
    use_tempfile: bool = False,
    suffix: str = "",
) -> bytes:
    """Call the command as a subprocess and return it's output as bytes.

    Args:
        data: The data to pass to the subprocess
        cmd: The command and arguments to call
        use_tempfile: If True, the command saves its output to a file, not stdout
        suffix: Suffix for the temporary file name

    Returns:
        The data printed to standard out by the subprocess.

    """
    # Convert all command arguments to strings
    cmd = list(map(str, cmd))

    # Convert data to bytes
    if isinstance(data, str):
        data = data.encode()

    if use_tempfile:
        # If the command cannot read from stdin, create a temporary file to pass to
        # the command
        with tempfile.NamedTemporaryFile(delete=False, suffix=suffix) as tfile:
            tfile.write(data)
        cmd.append(tfile.name)
        stdinput = None
    else:
        stdinput = data

    if log.level <= 0:
        import shlex

        log.debug("Running external command `%s`", shlex.join(cmd))

    error: Exception | None = None
    try:
        proc = await asyncio.create_subprocess_exec(
            *cmd,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.DEVNULL,
        )
        output_bytes, _ = await proc.communicate(stdinput)
    except FileNotFoundError as error:
        log.error("Could not run external command `%s`", cmd)
        raise error
    except subprocess.CalledProcessError as error:
        log.error("There was an error while running external command `%s`", cmd)
        raise error
    else:
        if (proc.returncode or 0) > 0 or error:
            # Raise an exception if the process failed so we can continue on the the
            # next conversion method
            raise subprocess.CalledProcessError(proc.returncode or 0, cmd)
    finally:
        # Clean up any temporary file
        if use_tempfile:
            tfile.close()
            Path(tfile.name).unlink()

    return output_bytes



"""Contain functions which convert data to formatted text."""

from __future__ import annotations

import logging
from functools import partial
from typing import TYPE_CHECKING

from prompt_toolkit.cache import SimpleCache
from prompt_toolkit.formatted_text import to_formatted_text

from euporie.core.convert.registry import register
from euporie.core.ft.ansi import ANSI
from euporie.core.ft.utils import strip_one_trailing_newline
from euporie.core.lexers import detect_lexer

if TYPE_CHECKING:
    from prompt_toolkit.formatted_text.base import StyleAndTextTuples

    from euporie.core.convert.datum import Datum
    from euporie.core.ft.html import HTML

log = logging.getLogger(__name__)

_html_cache: SimpleCache[str, HTML] = SimpleCache(maxsize=20)


@register(
    from_="html",
    to="ft",
)
async def html_to_ft(
    datum: Datum,
    cols: int | None = None,
    rows: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    extend: bool = True,
) -> StyleAndTextTuples:
    """Convert HTML to formatted text."""
    from euporie.core.ft.html import HTML

    data = datum.data
    markup = data.decode() if isinstance(data, bytes) else data
    html = _html_cache.get(
        datum.hash,
        partial(
            HTML,
            markup,
            width=cols,
            base=datum.path,
            collapse_root_margin=True,
            fill=extend,
            _initial_format=datum.root.format,
        ),
    )
    return await html._render(cols, rows)


_WHITELISTED_LEXERS = {
    "python",
    "markdown",
    "javascript",
    "json",
}


@register(
    from_="ansi",
    to="ft",
)
async def ansi_to_ft(
    datum: Datum,
    cols: int | None = None,
    rows: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    extend: bool = True,
    lex: bool = False,
) -> StyleAndTextTuples:
    """Convert ANSI text to formatted text, lexing & formatting automatically."""
    data = datum.data
    markup = data.decode() if isinstance(data, bytes) else data
    ft: StyleAndTextTuples
    if "\x1b" in markup or "\r" in markup:
        ft = to_formatted_text(ANSI(markup.strip()))
    else:
        # Replace tabs with spaces
        markup = markup.expandtabs()
        # Use lexer whitelist
        if (
            lex
            and (lexer := detect_lexer(markup, path=datum.path)) is not None
            and lexer.name in _WHITELISTED_LEXERS
        ):
            from prompt_toolkit.lexers.pygments import _token_cache

            log.debug('Lexing output using "%s" lexer', lexer.name)
            ft = [
                (_token_cache[t], v) for _, t, v in lexer.get_tokens_unprocessed(markup)
            ]

        else:
            ft = to_formatted_text(markup)
    return to_formatted_text(strip_one_trailing_newline(ft))



"""Key bindings to deal with pixel mouse positioning."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, NamedTuple

from prompt_toolkit.cache import FastDictCache
from prompt_toolkit.data_structures import Point
from prompt_toolkit.key_binding.bindings.mouse import (
    MOUSE_MOVE,
    UNKNOWN_BUTTON,
    UNKNOWN_MODIFIER,
    KeyBindings,
    typical_mouse_events,
    urxvt_mouse_events,
    xterm_sgr_mouse_events,
)
from prompt_toolkit.key_binding.bindings.mouse import (
    load_mouse_bindings as load_ptk_mouse_bindings,
)
from prompt_toolkit.keys import Keys
from prompt_toolkit.mouse_events import MouseButton, MouseEventType, MouseModifier
from prompt_toolkit.mouse_events import MouseEvent as PtkMouseEvent
from prompt_toolkit.renderer import HeightIsUnknownError

from euporie.core.app.app import BaseApp

if TYPE_CHECKING:
    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent

log = logging.getLogger(__name__)


class RelativePosition(NamedTuple):
    """Store the relative position or the mouse within a terminal cell."""

    x: float
    y: float


class MouseEvent(PtkMouseEvent):
    """Mouse event, which also store relative position of the mouse event in a cell."""

    def __init__(
        self,
        position: Point,
        event_type: MouseEventType,
        button: MouseButton,
        modifiers: frozenset[MouseModifier],
        cell_position: RelativePosition | None,
    ) -> None:
        """Create new event instance."""
        super().__init__(
            position=position,
            event_type=event_type,
            button=button,
            modifiers=modifiers,
        )
        self.cell_position = cell_position or RelativePosition(0.5, 0.5)


def _parse_mouse_data(
    data: str, sgr_pixels: bool, cell_size_xy: tuple[int, int]
) -> MouseEvent | None:
    """Convert key-press data to a mouse event."""
    rx = ry = 0.5

    # Parse incoming packet.
    if data[2] == "M":
        # Typical.
        mouse_event, x, y = map(ord, data[3:])

        # TODO: Is it possible to add modifiers here?
        mouse_button, mouse_event_type, mouse_modifiers = typical_mouse_events[
            mouse_event
        ]

        # Handle situations where `PosixStdinReader` used surrogateescapes.
        if x >= 0xDC00:
            x -= 0xDC00
        if y >= 0xDC00:
            y -= 0xDC00

        x -= 32
        y -= 32
    else:
        # Urxvt and Xterm SGR.
        # When the '<' is not present, we are not using the Xterm SGR mode,
        # but Urxvt instead.
        data = data[2:]
        if data[:1] == "<":
            sgr = True
            data = data[1:]
        else:
            sgr = False

        # Extract coordinates.
        mouse_event, x, y = map(int, data[:-1].split(";"))
        m = data[-1]

        # Parse event type.
        if sgr:
            if sgr_pixels:
                # Scale down pixel-wise mouse position to cell based, and calculate
                # relative position of mouse within the cell
                cell_x, cell_y = cell_size_xy
                px, py = x, y
                fx, fy = px / cell_x + 1, py / cell_y + 1
                x, y = int(fx), int(fy)
                rx, ry = fx - x, fy - y
            try:
                (mouse_button, mouse_event_type, mouse_modifiers) = (
                    xterm_sgr_mouse_events[mouse_event, m]
                )
            except KeyError:
                return None

        else:
            # Some other terminals, like urxvt, Hyper terminal, ...
            (mouse_button, mouse_event_type, mouse_modifiers) = urxvt_mouse_events.get(
                mouse_event, (UNKNOWN_BUTTON, MOUSE_MOVE, UNKNOWN_MODIFIER)
            )

    x -= 1
    y -= 1

    return MouseEvent(
        position=Point(x=x, y=y),
        event_type=mouse_event_type,
        button=mouse_button,
        modifiers=mouse_modifiers,
        cell_position=RelativePosition(rx, ry),
    )


_MOUSE_EVENT_CACHE: FastDictCache[
    tuple[str, bool, tuple[int, int]], MouseEvent | None
] = FastDictCache(get_value=_parse_mouse_data)


def load_mouse_bindings() -> KeyBindings:
    """Additional key-bindings to deal with SGR-pixel mouse positioning."""
    key_bindings = load_ptk_mouse_bindings()

    @key_bindings.add(Keys.Vt100MouseEvent, eager=True)
    def _(event: KeyPressEvent) -> NotImplementedOrNone:
        """Handle incoming mouse event, include SGR-pixel mode."""
        # Ensure mypy knows this would only run in a euporie appo
        app = event.app
        assert isinstance(app, BaseApp)

        if not app.renderer.height_is_known:
            return NotImplemented

        mouse_event = _MOUSE_EVENT_CACHE[
            event.data, app.term_sgr_pixel, app.cell_size_px
        ]

        if mouse_event is None:
            return NotImplemented

        # Only handle mouse events when we know the window height.
        if mouse_event.event_type is not None:
            # Take region above the layout into account. The reported
            # coordinates are absolute to the visible part of the terminal.
            x, y = mouse_event.position

            # Adjust position to take into account space above non-full screen apps
            try:
                rows_above = app.renderer.rows_above_layout
            except HeightIsUnknownError:
                return NotImplemented
            else:
                y -= rows_above

            # Save mouse position within the app
            app.mouse_position = Point(x=x, y=y)

            # Apply limits to mouse position if enabled
            if (mouse_limits := app.mouse_limits) is not None:
                x = max(
                    mouse_limits.xpos,
                    min(x, mouse_limits.xpos + (mouse_limits.width) - 1),
                )
                y = max(
                    mouse_limits.ypos,
                    min(y, mouse_limits.ypos + (mouse_limits.height) - 1),
                )

            # Do not modify the mouse event in the cache, instead create a new instance
            mouse_event = MouseEvent(
                position=Point(x=x, y=y),
                event_type=mouse_event.event_type,
                button=mouse_event.button,
                modifiers=mouse_event.modifiers,
                cell_position=mouse_event.cell_position,
            )

            # Call the mouse handler from the renderer.
            # Note: This can return `NotImplemented` if no mouse handler was
            #       found for this position, or if no repainting needs to
            #       happen. this way, we avoid excessive repaints during mouse
            #       movements.
            handler = app.renderer.mouse_handlers.mouse_handlers[y][x]

            return handler(mouse_event)

        return NotImplemented

    return key_bindings



"""Miscellaneous control fields."""

from __future__ import annotations

from typing import TYPE_CHECKING

from prompt_toolkit.layout.controls import UIContent, UIControl

if TYPE_CHECKING:
    from prompt_toolkit.formatted_text import (
        StyleAndTextTuples,
    )


class DummyControl(UIControl):
    """A dummy control object that doesn't paint any content."""

    def create_content(self, width: int, height: int) -> UIContent:
        """Return one blank line only."""

        def get_line(i: int) -> StyleAndTextTuples:
            return []

        return UIContent(get_line=get_line, line_count=1)


class FocusableDummyControl(DummyControl):
    """A dummy control object that doesn't paint any content."""

    def is_focusable(self) -> bool:
        """Make this control focusable."""
        return True



"""Define page navigation key-bindings for buffers."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from prompt_toolkit.application.current import get_app
from prompt_toolkit.filters import buffer_has_focus
from prompt_toolkit.key_binding.bindings.page_navigation import (
    load_emacs_page_navigation_bindings,
    load_vi_page_navigation_bindings,
)
from prompt_toolkit.key_binding.key_bindings import (
    ConditionalKeyBindings,
    merge_key_bindings,
)

from euporie.core.commands import add_cmd
from euporie.core.filters import micro_mode
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)

if TYPE_CHECKING:
    from prompt_toolkit.key_binding import KeyBindingsBase, KeyPressEvent

    from euporie.core.config import Config

log = logging.getLogger(__name__)


class PageNavigation:
    """Key-bindings for page navigation."""

    # Register default bindings for micro edit mode
    register_bindings(
        {
            "euporie.core.key_binding.bindings.page_navigation.PageNavigation": {
                "scroll-page-up": "pageup",
                "scroll-page-down": "pagedown",
            },
        }
    )


def load_page_navigation_bindings(config: Config | None = None) -> KeyBindingsBase:
    """Load page navigation key-bindings for text entry."""
    return ConditionalKeyBindings(
        merge_key_bindings(
            [
                load_emacs_page_navigation_bindings(),
                load_vi_page_navigation_bindings(),
                ConditionalKeyBindings(
                    load_registered_bindings(
                        "euporie.core.key_binding.bindings.page_navigation.PageNavigation",
                        config=config,
                    ),
                    micro_mode,
                ),
            ]
        ),
        buffer_has_focus,
    )


# Commands


@add_cmd(filter=buffer_has_focus, hidden=True)
def scroll_page_down(event: KeyPressEvent) -> None:
    """Scroll page down (prefer the cursor at the top of the page, after scrolling)."""
    w = event.app.layout.current_window
    b = event.app.current_buffer

    if w and w.render_info:
        # Scroll down one page.
        line_index = b.document.cursor_position_row
        page = w.render_info.window_height

        if (
            (screen := get_app().renderer._last_screen)
            and (wp := screen.visible_windows_to_write_positions.get(w))
            and (bbox := getattr(wp, "bbox", None))
        ):
            page -= bbox.top + bbox.bottom

        line_index = max(line_index + page, w.vertical_scroll + 1)
        w.vertical_scroll = line_index

        b.cursor_position = b.document.translate_row_col_to_index(line_index, 0)
        b.cursor_position += b.document.get_start_of_line_position(
            after_whitespace=True
        )


@add_cmd(filter=buffer_has_focus, hidden=True)
def scroll_page_up(event: KeyPressEvent) -> None:
    """Scroll page up (prefer the cursor at the bottom of the page, after scrolling)."""
    w = event.app.layout.current_window
    b = event.app.current_buffer

    if w and w.render_info:
        line_index = b.document.cursor_position_row
        page = w.render_info.window_height

        if (
            (screen := get_app().renderer._last_screen)
            and (wp := screen.visible_windows_to_write_positions.get(w))
            and (bbox := getattr(wp, "bbox", None))
        ):
            page -= bbox.top + bbox.bottom

        # Put cursor at the first visible line. (But make sure that the cursor
        # moves at least one line up.)
        line_index = max(
            0,
            min(line_index - page, b.document.cursor_position_row - 1),
        )

        b.cursor_position = b.document.translate_row_col_to_index(line_index, 0)
        b.cursor_position += b.document.get_start_of_line_position(
            after_whitespace=True
        )

        # Set the scroll offset. We can safely set it to zero; the Window will
        # make sure that it scrolls at least until the cursor becomes visible.
        w.vertical_scroll = 0



"""Contain function which convert data to rich format."""

from __future__ import annotations

from typing import TYPE_CHECKING

from euporie.core.convert.registry import register
from euporie.core.filters import have_modules

if TYPE_CHECKING:
    from rich.markdown import Markdown

    from euporie.core.convert.datum import Datum


@register(
    from_="markdown",
    to="rich",
    filter_=have_modules("rich"),
)
async def markdown_to_rich_py(
    datum: Datum,
    cols: int | None = None,
    rows: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    extend: bool = True,
) -> Markdown:
    """Convert base64 encoded data to bytes."""
    from rich.markdown import Markdown

    data = datum.data
    markup = data.decode() if isinstance(data, bytes) else data
    return Markdown(
        markup,
        # code_theme=str(config.syntax_theme),
        # inline_code_theme=str(config.syntax_theme),
    )



"""Concern dumping output."""

from __future__ import annotations

import io
import logging
import os
import sys
from functools import partial
from typing import TYPE_CHECKING, cast

from prompt_toolkit.layout.containers import DynamicContainer, FloatContainer, Window
from prompt_toolkit.output.defaults import create_output
from prompt_toolkit.output.vt100 import Vt100_Output

from euporie.core.app.app import BaseApp
from euporie.core.app.current import get_app
from euporie.core.io import PseudoTTY
from euporie.core.key_binding.registry import register_bindings
from euporie.preview.tabs.notebook import PreviewNotebook

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any, TextIO

    from prompt_toolkit.application.application import _AppResult
    from prompt_toolkit.layout.containers import Float
    from prompt_toolkit.output import Output

    from euporie.core.tabs.base import Tab

log = logging.getLogger(__name__)


class PreviewApp(BaseApp):
    """Preview app.

    Preview notebook files in the terminal.

    Outputs a formatted notebook file. The formatted output will be written to
    the the output file path given by `output_file` (the standard output by
    default).

    """

    name = "preview"

    def __init__(self, **kwargs: Any) -> None:
        """Create an app for dumping a prompt-toolkit layout."""
        # Set default arguments
        kwargs.setdefault("title", "euporie-preview")
        kwargs.setdefault("leave_graphics", True)
        kwargs.setdefault("full_screen", False)
        kwargs.setdefault("max_render_postpone_time", 0)
        kwargs.setdefault("min_redraw_interval", 0)
        kwargs.setdefault("extend_renderer_height", True)
        kwargs.setdefault("enable_page_navigation_bindings", False)
        # Adjust options if we are paging output
        if self.config.page:
            kwargs.setdefault("set_title", False)
            kwargs.setdefault("extend_renderer_width", True)
        # Initialise the application
        super().__init__(**kwargs)
        # We want the app to close when rendering is complete
        # self.after_render += self.pre_exit
        # Do not load any key bindings
        self.bindings_to_load.append("euporie.preview.app.PreviewApp")
        # Select the first tab after files are opened
        self.pre_run_callables.append(partial(setattr, self, "tab_idx", 0))

    def get_file_tab(self, path: Path) -> type[Tab]:
        """Return the tab to use for a file path."""
        return PreviewNotebook

    def exit(
        self,
        result: _AppResult | None = None,
        exception: BaseException | type[BaseException] | None = None,
        style: str = "",
    ) -> None:
        """Optionally pipe the output to a pager on exit."""
        # Display pager if needed
        if self.config.page:
            from pydoc import pager

            output_file = getattr(self.output, "output_file")  # noqa: B009
            if output_file is not None:
                output_file.seek(0)
                data = output_file.read()
                pager(data)
        if exception is not None:
            super().exit(exception=exception, style=style)
        elif result is not None:
            super().exit(result=result, style=style)
        else:
            super().exit()

    def load_container(self) -> FloatContainer:
        """Return a container with all opened tabs."""
        return FloatContainer(
            DynamicContainer(lambda: self.tab or Window()),
            floats=cast("list[Float]", self.floats),
        )

    def cleanup_closed_tab(self, tab: Tab) -> None:
        """Exit if all tabs are closed."""
        super().cleanup_closed_tab(tab)
        if not self.tabs:
            self._is_running = False
            self.exit()
        self.draw(render_as_done=True)

    @classmethod
    def load_output(cls) -> Output:
        """Load the output.

        Depending on the application configuration, will set the output to a file, to
        stdout, or to a temporary file so the output can be displayed in a pager.

        Returns:
            A container for notebook output

        """
        output_file: TextIO
        if cls.config.page:
            # Use a temporary file as display output if we are going to page the output
            from tempfile import TemporaryFile

            output_file = TemporaryFile("w+")  # noqa: SIM115
            # Make this file look like a tty so we get colorful output
            output_file = cast("TextIO", PseudoTTY(output_file, isatty=True))

        else:
            # If we are not paging output, determine where to print it
            if cls.config.output_file is None or str(cls.config.output_file) in (
                "-",
                "/dev/stdout",
            ):
                output_file = sys.stdout
            elif str(cls.config.output_file) == "/dev/stderr":
                output_file = sys.stderr
            else:
                try:
                    output_file = cls.config.output_file.open("w+")
                except (
                    FileNotFoundError,
                    PermissionError,
                    io.UnsupportedOperation,
                ) as error:
                    log.error(error)
                    log.error(
                        "Output file `%s` cannot be opened. "
                        "Standard output will be used.",
                        cls.config.output_file,
                    )
                    output_file = sys.stdout

            # Make the output look like a TTY if color-depth has been configured
            if not output_file.isatty() and cls.config.color_depth is not None:
                output_file = cast(
                    "TextIO",
                    PseudoTTY(
                        output_file,
                        isatty=True,
                    ),
                )

        # Ensure we do not receive the "Output is not a terminal" message
        Vt100_Output._fds_not_a_terminal.add(output_file.fileno())
        # Set environment variable to disable character position requests
        os.environ["PROMPT_TOOLKIT_NO_CPR"] = "1"
        # Create a default output - this detects the terminal type
        # Do not use stderr instead of stdout if stdout is not a tty
        output = create_output(cast("TextIO", output_file), always_prefer_tty=False)

        # Use the width and height of stderr (this gives us the terminal size even if
        # output is being piped to a non-tty)
        setattr(  # noqa B010
            output, "get_size", create_output(stdout=sys.stderr).get_size
        )
        # Attach the output file to the output in case we need to page it
        setattr(output, "output_file", output_file)  # noqa B010

        return output

    def _redraw(self, render_as_done: bool = False) -> None:
        """Ensure the output is always rendered as done."""
        # import time
        # time.sleep(0.1)
        super()._redraw(render_as_done=True)

    def _update_invalidate_events(self) -> None:
        """Do nothing, as we don't need invalidation events for the preview app."""

    # def load_key_bindings(self) -> None:
    #     """Do no load any additional key-bindings for the preview app."""

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.preview.app.PreviewApp": {
                "quit": ["c-c", "c-q"],
            }
        }
    )


def get_preview_app() -> PreviewApp:
    """Get the current application."""
    return cast("PreviewApp", get_app())



"""Main entry point into euporie.core."""

from __future__ import annotations

from functools import cache
from importlib.metadata import entry_points
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from importlib.metadata import EntryPoint, EntryPoints


@cache
def available_apps() -> dict[str, EntryPoint]:
    """Return a list of loadable euporie apps."""
    eps: dict | EntryPoints
    try:
        eps = entry_points(group="euporie.apps")
    except TypeError:
        eps = entry_points()
    if isinstance(eps, dict):
        points = eps.get("euporie.apps")
    else:
        points = eps.select(group="euporie.apps")
    apps = {x.name: x for x in points} if points else {}
    return apps


def main(name: str = "launch") -> None:
    """Load and launches the application."""
    # Register extensions to external packages
    from euporie.core import (
        path,  # noqa F401
        pygments,  # noqa F401
    )

    # Monkey-patch prompt_toolkit
    from euporie.core.layout import containers  # noqa: F401

    apps = available_apps()
    if entry := apps.get(name):
        return entry.load().launch()
    else:
        raise ModuleNotFoundError(f"Euporie app `{name}` not installed")


if __name__ == "__main__":
    main()



"""Commands for the notebook app."""

from euporie.core.commands import add_cmd


@add_cmd(aliases=["n"])
def _new_notebook() -> None:
    """Create a new file."""
    from euporie.notebook.current import get_app
    from euporie.notebook.tabs.notebook import Notebook

    app = get_app()
    app.add_tab(tab := Notebook(app, None))
    tab.focus()


@add_cmd()
def _view_documentation() -> None:
    """Open the documentation in a web-view tab."""
    from upath import UPath

    from euporie.core.app.current import get_app

    get_app().open_file(UPath("https://euporie.readthedocs.io/"))
